# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from enum import IntEnum, auto
from typing import Dict, List


class Context(IntEnum):
    PROGRAM = auto()
    PROJECT = auto()
    OVERVIEW_TAB = auto()
    DATA_SETS_TAB = auto()
    KRAMERS_KRONIG_TAB = auto()
    DRT_TAB = auto()
    FITTING_TAB = auto()
    SIMULATION_TAB = auto()
    PLOTTING_TAB = auto()


class Action(IntEnum):
    # Program-level
    NEW_PROJECT = auto()
    LOAD_PROJECT = auto()
    EXIT = auto()
    NEXT_PROGRAM_TAB = auto()
    PREVIOUS_PROGRAM_TAB = auto()
    SELECT_HOME_TAB = auto()
    SHOW_HELP_ABOUT = auto()
    SHOW_HELP_LICENSES = auto()
    SHOW_SETTINGS_APPEARANCE = auto()
    SHOW_SETTINGS_DEFAULTS = auto()
    SHOW_SETTINGS_KEYBINDINGS = auto()
    SHOW_COMMAND_PALETTE = auto()
    SHOW_CHANGELOG = auto()
    CHECK_UPDATES = auto()

    # Project-level
    SAVE_PROJECT = auto()
    SAVE_PROJECT_AS = auto()
    CLOSE_PROJECT = auto()
    UNDO = auto()
    REDO = auto()
    NEXT_PROJECT_TAB = auto()
    PREVIOUS_PROJECT_TAB = auto()
    SELECT_DATA_SETS_TAB = auto()
    SELECT_DRT_TAB = auto()
    SELECT_FITTING_TAB = auto()
    SELECT_KRAMERS_KRONIG_TAB = auto()
    SELECT_OVERVIEW_TAB = auto()
    SELECT_PLOTTING_TAB = auto()
    SELECT_SIMULATION_TAB = auto()

    # Project-level: multiple tabs
    PERFORM_ACTION = auto()
    # - Load data set
    # - Perform test
    # - Perform fit
    # - Perform simulation
    # - Create plot
    DELETE_RESULT = auto()
    # - Data set
    # - Test result
    # - Fit result
    # - Simulation result
    # - Plot
    NEXT_PRIMARY_RESULT = auto()
    PREVIOUS_PRIMARY_RESULT = auto()
    # - Data set
    # - Plot
    NEXT_SECONDARY_RESULT = auto()
    PREVIOUS_SECONDARY_RESULT = auto()
    # - Test result
    # - Fit result
    # - Simulation result
    # - Plot type
    APPLY_SETTINGS = auto()
    # - Kramers-Kronig tab
    # - DRT tab
    # - Fitting tab
    # - Simulation tab
    APPLY_MASK = auto()
    # - Kramers-Kronig tab
    # - DRT tab
    # - Fitting tab
    SHOW_ENLARGED_NYQUIST = auto()
    SHOW_ENLARGED_BODE = auto()
    SHOW_ENLARGED_RESIDUALS = auto()
    SHOW_ENLARGED_DRT = auto()
    SHOW_ENLARGED_IMPEDANCE = auto()
    SHOW_CIRCUIT_EDITOR = auto()
    # - Fitting tab
    # - Simulation tab
    COPY_DRT_DATA = auto()
    COPY_IMPEDANCE_DATA = auto()
    COPY_NYQUIST_DATA = auto()
    COPY_BODE_DATA = auto()
    COPY_RESIDUALS_DATA = auto()
    COPY_OUTPUT = auto()
    # - Fit output
    # - Simulation output
    LOAD_SIMULATION_AS_DATA_SET = auto()

    # Project-level: data sets tab
    AVERAGE_DATA_SETS = auto()
    TOGGLE_DATA_POINTS = auto()
    COPY_DATA_SET_MASK = auto()
    SUBTRACT_IMPEDANCE = auto()

    # Project-level: plotting tab
    SELECT_ALL_PLOT_SERIES = auto()
    UNSELECT_ALL_PLOT_SERIES = auto()
    COPY_PLOT_APPEARANCE = auto()
    COPY_PLOT_DATA = auto()
    EXPAND_COLLAPSE_SIDEBAR = auto()
    EXPORT_PLOT = auto()


action_contexts: Dict[Action, List[Context]] = {
    Action.NEW_PROJECT: [Context.PROGRAM],
    Action.LOAD_PROJECT: [Context.PROGRAM],
    Action.EXIT: [Context.PROGRAM],
    Action.NEXT_PROGRAM_TAB: [Context.PROGRAM],
    Action.PREVIOUS_PROGRAM_TAB: [Context.PROGRAM],
    Action.SELECT_HOME_TAB: [Context.PROGRAM],
    Action.SHOW_HELP_ABOUT: [Context.PROGRAM],
    Action.SHOW_HELP_LICENSES: [Context.PROGRAM],
    Action.SHOW_SETTINGS_APPEARANCE: [Context.PROGRAM],
    Action.SHOW_SETTINGS_DEFAULTS: [Context.PROGRAM],
    Action.SHOW_SETTINGS_KEYBINDINGS: [Context.PROGRAM],
    Action.SHOW_COMMAND_PALETTE: [Context.PROGRAM],
    Action.SHOW_CHANGELOG: [Context.PROGRAM],
    Action.CHECK_UPDATES: [Context.PROGRAM],
    Action.SAVE_PROJECT: [Context.PROJECT],
    Action.SAVE_PROJECT_AS: [Context.PROJECT],
    Action.CLOSE_PROJECT: [Context.PROJECT],
    Action.UNDO: [Context.PROJECT],
    Action.REDO: [Context.PROJECT],
    Action.NEXT_PROJECT_TAB: [Context.PROJECT],
    Action.PREVIOUS_PROJECT_TAB: [Context.PROJECT],
    Action.SELECT_DATA_SETS_TAB: [Context.PROJECT],
    Action.SELECT_DRT_TAB: [Context.PROJECT],
    Action.SELECT_FITTING_TAB: [Context.PROJECT],
    Action.SELECT_KRAMERS_KRONIG_TAB: [Context.PROJECT],
    Action.SELECT_OVERVIEW_TAB: [Context.PROJECT],
    Action.SELECT_PLOTTING_TAB: [Context.PROJECT],
    Action.SELECT_SIMULATION_TAB: [Context.PROJECT],
    Action.PERFORM_ACTION: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.DELETE_RESULT: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.NEXT_PRIMARY_RESULT: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.PREVIOUS_PRIMARY_RESULT: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.NEXT_SECONDARY_RESULT: [
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.PREVIOUS_SECONDARY_RESULT: [
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.APPLY_SETTINGS: [
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.APPLY_MASK: [
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
    ],
    Action.SHOW_ENLARGED_IMPEDANCE: [
        Context.DRT_TAB,
    ],
    Action.SHOW_ENLARGED_DRT: [
        Context.DRT_TAB,
    ],
    Action.SHOW_ENLARGED_NYQUIST: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.SHOW_ENLARGED_BODE: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.SHOW_ENLARGED_RESIDUALS: [
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
    ],
    Action.SHOW_CIRCUIT_EDITOR: [
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.COPY_DRT_DATA: [
        Context.DRT_TAB,
    ],
    Action.COPY_IMPEDANCE_DATA: [
        Context.DRT_TAB,
    ],
    Action.COPY_NYQUIST_DATA: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.COPY_BODE_DATA: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.COPY_RESIDUALS_DATA: [
        Context.KRAMERS_KRONIG_TAB,
        Context.DRT_TAB,
        Context.FITTING_TAB,
    ],
    Action.COPY_OUTPUT: [
        Context.DRT_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.LOAD_SIMULATION_AS_DATA_SET: [Context.SIMULATION_TAB],
    Action.AVERAGE_DATA_SETS: [Context.DATA_SETS_TAB],
    Action.TOGGLE_DATA_POINTS: [Context.DATA_SETS_TAB],
    Action.COPY_DATA_SET_MASK: [Context.DATA_SETS_TAB],
    Action.SUBTRACT_IMPEDANCE: [Context.DATA_SETS_TAB],
    Action.SELECT_ALL_PLOT_SERIES: [Context.PLOTTING_TAB],
    Action.UNSELECT_ALL_PLOT_SERIES: [Context.PLOTTING_TAB],
    Action.COPY_PLOT_APPEARANCE: [Context.PLOTTING_TAB],
    Action.COPY_PLOT_DATA: [Context.PLOTTING_TAB],
    Action.EXPAND_COLLAPSE_SIDEBAR: [Context.PLOTTING_TAB],
    Action.EXPORT_PLOT: [Context.PLOTTING_TAB],
}


action_to_string: Dict[Action, str] = {
    Action.APPLY_MASK: "apply-mask",
    Action.APPLY_SETTINGS: "apply-settings",
    Action.AVERAGE_DATA_SETS: "average-data-sets",
    Action.CHECK_UPDATES: "check-updates",
    Action.CLOSE_PROJECT: "close-project",
    Action.COPY_BODE_DATA: "copy-bode-data",
    Action.COPY_DATA_SET_MASK: "copy-data-set-mask",
    Action.COPY_DRT_DATA: "copy-drt-data",
    Action.COPY_IMPEDANCE_DATA: "copy-impedance-data",
    Action.COPY_NYQUIST_DATA: "copy-nyquist-data",
    Action.COPY_OUTPUT: "copy-output",
    Action.COPY_PLOT_APPEARANCE: "copy-plot-appearance",
    Action.COPY_PLOT_DATA: "copy-plot-data",
    Action.COPY_RESIDUALS_DATA: "copy-residuals-data",
    Action.DELETE_RESULT: "delete-result",
    Action.EXIT: "exit-program",
    Action.EXPAND_COLLAPSE_SIDEBAR: "expand-collapse-sidebar",
    Action.EXPORT_PLOT: "export-plot",
    Action.LOAD_PROJECT: "load-project",
    Action.LOAD_SIMULATION_AS_DATA_SET: "load-simulation-as-data-set",
    Action.NEW_PROJECT: "new-project",
    Action.NEXT_PRIMARY_RESULT: "next-primary-result",
    Action.NEXT_PROGRAM_TAB: "next-program-tab",
    Action.NEXT_PROJECT_TAB: "next-project-tab",
    Action.NEXT_SECONDARY_RESULT: "next-secondary-result",
    Action.PERFORM_ACTION: "perform-action",
    Action.PREVIOUS_PRIMARY_RESULT: "previous-primary-result",
    Action.PREVIOUS_PROGRAM_TAB: "previous-program-tab",
    Action.PREVIOUS_PROJECT_TAB: "previous-project-tab",
    Action.PREVIOUS_SECONDARY_RESULT: "previous-secondary-result",
    Action.REDO: "redo",
    Action.SAVE_PROJECT: "save-project",
    Action.SAVE_PROJECT_AS: "save-project-as",
    Action.SELECT_ALL_PLOT_SERIES: "select-all-plot-series",
    Action.SELECT_DATA_SETS_TAB: "select-data-sets-tab",
    Action.SELECT_DRT_TAB: "select-drt-tab",
    Action.SELECT_FITTING_TAB: "select-fitting-tab",
    Action.SELECT_HOME_TAB: "select-home-tab",
    Action.SELECT_KRAMERS_KRONIG_TAB: "select-kramers-kronig-tab",
    Action.SELECT_OVERVIEW_TAB: "select-overview-tab",
    Action.SELECT_PLOTTING_TAB: "select-plotting-tab",
    Action.SELECT_SIMULATION_TAB: "select-simulation-tab",
    Action.SHOW_CHANGELOG: "show-changelog",
    Action.SHOW_CIRCUIT_EDITOR: "show-circuit-editor",
    Action.SHOW_COMMAND_PALETTE: "show-command-palette",
    Action.SHOW_ENLARGED_BODE: "show-enlarged-bode",
    Action.SHOW_ENLARGED_DRT: "show-enlarged-drt",
    Action.SHOW_ENLARGED_IMPEDANCE: "show-enlarged-impedance",
    Action.SHOW_ENLARGED_NYQUIST: "show-enlarged-nyquist",
    Action.SHOW_ENLARGED_RESIDUALS: "show-enlarged-residuals",
    Action.SHOW_HELP_ABOUT: "show-help-about",
    Action.SHOW_HELP_LICENSES: "show-help-licenses",
    Action.SHOW_SETTINGS_APPEARANCE: "show-settings-appearance",
    Action.SHOW_SETTINGS_DEFAULTS: "show-settings-defaults",
    Action.SHOW_SETTINGS_KEYBINDINGS: "show-settings-keybindings",
    Action.SUBTRACT_IMPEDANCE: "subtract-impedance",
    Action.TOGGLE_DATA_POINTS: "toggle-data-points",
    Action.UNDO: "undo",
    Action.UNSELECT_ALL_PLOT_SERIES: "unselect-all-plot-series",
}
string_to_action: Dict[str, Action] = {v: k for k, v in action_to_string.items()}
# Check that there are no duplicate keys
assert len(action_to_string) == len(set(action_to_string.values())) and len(
    action_to_string
) == len(string_to_action), "Duplicate action string keys detected!"


action_descriptions: Dict[Action, str] = {
    Action.NEW_PROJECT: """
Create a new project.
""".strip(),
    Action.LOAD_PROJECT: """
Select project(s) to load.
""".strip(),
    Action.EXIT: """
Close the program.
""".strip(),
    Action.NEXT_PROGRAM_TAB: """
Go to the next project.
""".strip(),
    Action.PREVIOUS_PROGRAM_TAB: """
Go to the previous project.
""".strip(),
    Action.SELECT_HOME_TAB: """
Go to the 'Home' tab.
""".strip(),
    Action.SHOW_HELP_ABOUT: """
Show the 'About' window.
""".strip(),
    Action.SHOW_HELP_LICENSES: """
Show the 'Licenses' window.
""".strip(),
    Action.SHOW_SETTINGS_APPEARANCE: """
Show the 'Settings - appearance' window.
""".strip(),
    Action.SHOW_SETTINGS_DEFAULTS: """
Show the 'Settings - defaults' window.
""".strip(),
    Action.SHOW_SETTINGS_KEYBINDINGS: """
Show the 'Settings - keybindings' window.
""".strip(),
    Action.SHOW_COMMAND_PALETTE: """
Show the command palette, which can be used as an alternative to other keybindings for performing actions.
""".strip(),
    Action.SHOW_CHANGELOG: """
Show the changelog.
""".strip(),
    Action.CHECK_UPDATES: """
Check for updates.
""".strip(),
    Action.SAVE_PROJECT: """
Save the current project.
""".strip(),
    Action.SAVE_PROJECT_AS: """
Save the current project as a new file.
""".strip(),
    Action.CLOSE_PROJECT: """
Close the current project.
""".strip(),
    Action.UNDO: """
Undo the latest action.
""".strip(),
    Action.REDO: """
Redo the latest action.
""".strip(),
    Action.NEXT_PROJECT_TAB: """
Go to the next project tab.
""".strip(),
    Action.PREVIOUS_PROJECT_TAB: """
Go to the previous project tab.
""".strip(),
    Action.SELECT_DATA_SETS_TAB: """
Go to the 'Data sets' tab.
""".strip(),
    Action.SELECT_DRT_TAB: """
Go to the 'DRT analysis' tab.
""".strip(),
    Action.SELECT_FITTING_TAB: """
Go to the 'Fitting' tab.
""".strip(),
    Action.SELECT_KRAMERS_KRONIG_TAB: """
Go to the 'Kramers-Kronig' tab.
""".strip(),
    Action.SELECT_OVERVIEW_TAB: """
Go to the 'Overview' tab.
""".strip(),
    Action.SELECT_PLOTTING_TAB: """
Go to the 'Plotting' tab.
""".strip(),
    Action.SELECT_SIMULATION_TAB: """
Go to the 'Simulation' tab.
""".strip(),
    Action.PERFORM_ACTION: """
Perform the primary action of the current project tab:
- Data sets: select files to load.
- Kramers-Kronig: perform test.
- DRT analysis: perform analysis.
- Fitting: perform fit.
- Simulation: perform simulation.
- Plotting: create a new plot.
""".strip(),
    Action.DELETE_RESULT: """
Delete the current result in the current project tab:
- Data sets: delete the current data set.
- Kramers-Kronig: delete the current test result.
- DRT analysis: delete the current analysis result.
- Fitting: delete the current fit result.
- Simulation: delete the current simulation result.
- Plotting: delete the current plot.
""".strip(),
    Action.NEXT_PRIMARY_RESULT: """
Select the next primary result of the current project tab:
- Data sets: data set.
- Kramers-Kronig: data set.
- DRT analysis: data set.
- Fitting: data set.
- Simulation: data set.
- Plotting: plot.
""".strip(),
    Action.PREVIOUS_PRIMARY_RESULT: """
Select the previous primary result of the current project tab:
- Data sets: data set.
- Kramers-Kronig: data set.
- DRT analysis: data set.
- Fitting: data set.
- Simulation: data set.
- Plotting: plot.
""".strip(),
    Action.NEXT_SECONDARY_RESULT: """
Select the next secondary result of the current project tab:
- Kramers-Kronig: test result.
- DRT analysis: analysis result.
- Fitting: fit result.
- Simulation: simulation result.
- Plotting: plot type.
""".strip(),
    Action.PREVIOUS_SECONDARY_RESULT: """
Select the previous secondary result of the current project tab:
- Kramers-Kronig: test result.
- DRT analysis: analysis result.
- Fitting: fit result.
- Simulation: simulation result.
- Plotting: plot type.
""".strip(),
    Action.APPLY_SETTINGS: """
Apply the settings used in the current secondary result of the current project tab:
- Kramers-Kronig: test result.
- DRT analysis: analysis result.
- Fitting: fit result.
- Simulation: simulation result.
""".strip(),
    Action.APPLY_MASK: """
Apply the mask used in the current secondary result of the current project tab:
- Kramers-Kronig: test result.
- DRT analysis: analysis result.
- Fitting: fit result.
""".strip(),
    Action.SHOW_ENLARGED_DRT: """
Show an enlarged DRT plot.
""".strip(),
    Action.SHOW_ENLARGED_IMPEDANCE: """
Show an enlarged impedance plot.
""".strip(),
    Action.SHOW_ENLARGED_NYQUIST: """
Show an enlarged Nyquist plot.
""".strip(),
    Action.SHOW_ENLARGED_BODE: """
Show an enlarged Bode plot.
""".strip(),
    Action.SHOW_ENLARGED_RESIDUALS: """
Show an enlarged residuals plot.
""".strip(),
    Action.SHOW_CIRCUIT_EDITOR: """
Show the circuit editor window.
""".strip(),
    Action.COPY_DRT_DATA: """
Copy the data from the DRT plot.
""".strip(),
    Action.COPY_IMPEDANCE_DATA: """
Copy the data from the impedance plot.
""".strip(),
    Action.COPY_NYQUIST_DATA: """
Copy the data from the Nyquist plot.
""".strip(),
    Action.COPY_BODE_DATA: """
Copy the data from the Bode plot.
""".strip(),
    Action.COPY_RESIDUALS_DATA: """
Copy the data from the residuals plot.
""".strip(),
    Action.COPY_OUTPUT: """
Copy the chosen output while in the 'DRT analysis', 'Fitting', or the 'Simulation' tab.
""".strip(),
    Action.AVERAGE_DATA_SETS: """
Select data sets to average.
""".strip(),
    Action.TOGGLE_DATA_POINTS: """
Select data points to toggle.
""".strip(),
    Action.COPY_DATA_SET_MASK: """
Select which data set's mask to copy.
""".strip(),
    Action.SUBTRACT_IMPEDANCE: """
Select the impedance to subtract from the current data set.
""".strip(),
    Action.SELECT_ALL_PLOT_SERIES: """
Select all data sets and test/fit/simulation results.
""".strip(),
    Action.UNSELECT_ALL_PLOT_SERIES: """
Unselect all data sets and test/fit/simulation results.
""".strip(),
    Action.COPY_PLOT_APPEARANCE: """
Select which plot's appearance settings to copy.
""".strip(),
    Action.COPY_PLOT_DATA: """
Copy the data from the current plot while in the 'Plotting' tab.
""".strip(),
    Action.EXPAND_COLLAPSE_SIDEBAR: """
Collapse/expand the sidebar while in the 'Plotting' tab.
""".strip(),
    Action.EXPORT_PLOT: """
Export the current plot using matplotlib.
""".strip(),
    Action.LOAD_SIMULATION_AS_DATA_SET: """
Load the current simulation as a data set.
""".strip(),
}
# Check that every action has a description
assert set(action_to_string.keys()) == set(
    action_descriptions.keys()
), "Missing action descriptions detected!"


class CNLSMethod(IntEnum):
    """
    Iterative methods used during complex non-linear least-squares fitting:

    - AUTO: try each method
    - AMPGO
    - BASINHOPPING
    - BFGS
    - BRUTE
    - CG
    - COBYLA
    - DIFFERENTIAL_EVOLUTION
    - DOGLEG
    - DUAL_ANNEALING
    - EMCEE
    - LBFGSB
    - LEASTSQ
    - LEAST_SQUARES
    - NELDER
    - NEWTON
    - POWELL
    - SHGO
    - SLSQP
    - TNC
    - TRUST_CONSTR
    - TRUST_EXACT
    - TRUST_KRYLOV
    - TRUST_NCG
    """

    AUTO = 1
    LEASTSQ = 2
    LEAST_SQUARES = 3
    DIFFERENTIAL_EVOLUTION = 4
    BRUTE = 5
    BASINHOPPING = 6
    AMPGO = 7
    NELDER = 8
    LBFGSB = 9
    POWELL = 10
    CG = 11
    NEWTON = 12
    COBYLA = 13
    BFGS = 14
    TNC = 15
    TRUST_NCG = 16
    TRUST_EXACT = 17
    TRUST_KRYLOV = 18
    TRUST_CONSTR = 19
    DOGLEG = 20
    SLSQP = 21
    EMCEE = 22
    SHGO = 23
    DUAL_ANNEALING = 24


label_to_cnls_method: Dict[str, CNLSMethod] = {
    "Auto": CNLSMethod.AUTO,
    "Levenberg-Marquardt": CNLSMethod.LEASTSQ,
    "Least-squares (trust-region reflective)": CNLSMethod.LEAST_SQUARES,
    "Powell": CNLSMethod.POWELL,  # Had some problem fitting
    "BFGS": CNLSMethod.BFGS,  # Had some problem fitting
    "Nelder-Mead": CNLSMethod.NELDER,  # Had some problem fitting
    "Conjugate Gradient": CNLSMethod.CG,  # Had some problem fitting
    "Truncated Newton": CNLSMethod.TNC,  # Had some problem fitting
    "L-BFGS-B": CNLSMethod.LBFGSB,  # NaN errors
    # "COBYLA": CNLSMethod.COBYLA,  # Had some problem fitting
    "Sequential Linear Squares Programming": CNLSMethod.SLSQP,  # Had some problem fitting
    # "Basin hopping": CNLSMethod.BASINHOPPING,  # NaN errors
    # "Differential Evolution": CNLSMethod.DIFFERENTIAL_EVOLUTION,  # Requires finite bounds for all mutable parameters
    # "Brute force method": CNLSMethod.BRUTE,  # Requires that brute_step is defined for mutable parameters
    # "Adaptive Memory Programming for Global Optimization": CNLSMethod.AMPGO,  # NaN errors
    # "Newton": CNLSMethod.NEWTON,  # Requires Jacobian
    # "Newton CG": CNLSMethod.TRUST_NCG,  # Requires Jacobian for trust region
    # "Exact trust-region": CNLSMethod.TRUST_EXACT,  # Requires Jacobian for trust region
    # "Newton GLTR trust-region": CNLSMethod.TRUST_KRYLOV,  # Requires Jacobian for trust region
    # "Constrained trust-region": CNLSMethod.TRUST_CONSTR,  # NaN errors
    # "Dog-leg trust-region": CNLSMethod.DOGLEG,  # Requires Jacobian
    # "Simplicial Homology Global Optimization": CNLSMethod.SHGO,  # NaN errors
    # "Dual Annealing": CNLSMethod.DUAL_ANNEALING,  # Requires finite bounds for all mutable parameters
    # "Maximum likelyhood via Monte-Carlo Markov chain": CNLSMethod.EMCEE,  # Requires the emcee package (version 3)
}
cnls_method_to_label: Dict[CNLSMethod, str] = {
    v: k for k, v in label_to_cnls_method.items()
}
cnls_method_to_label[CNLSMethod.AUTO] = "Auto"
cnls_method_to_value: Dict[CNLSMethod, str] = {
    CNLSMethod.AUTO: "auto",
    CNLSMethod.LEASTSQ: "leastsq",
    CNLSMethod.LEAST_SQUARES: "least_squares",
    CNLSMethod.DIFFERENTIAL_EVOLUTION: "differential_evolution",
    CNLSMethod.BRUTE: "brute",
    CNLSMethod.BASINHOPPING: "basinhopping",
    CNLSMethod.AMPGO: "ampgo",
    CNLSMethod.NELDER: "nelder",
    CNLSMethod.LBFGSB: "lbfgsb",
    CNLSMethod.POWELL: "powell",
    CNLSMethod.CG: "cg",
    CNLSMethod.NEWTON: "newton",
    CNLSMethod.COBYLA: "cobyla",
    CNLSMethod.BFGS: "bfgs",
    CNLSMethod.TNC: "tnc",
    CNLSMethod.TRUST_NCG: "trust-ncg",
    CNLSMethod.TRUST_EXACT: "trust-exact",
    CNLSMethod.TRUST_KRYLOV: "trust-krylov",
    CNLSMethod.TRUST_CONSTR: "trust-constr",
    CNLSMethod.DOGLEG: "dogleg",
    CNLSMethod.SLSQP: "slsqp",
    CNLSMethod.EMCEE: "emcee",
    CNLSMethod.SHGO: "shgo",
    CNLSMethod.DUAL_ANNEALING: "dual_annealing",
}
value_to_cnls_method: Dict[str, CNLSMethod] = {
    v: k for k, v in cnls_method_to_value.items()
}
assert set(cnls_method_to_value.keys()) == set(
    value_to_cnls_method.values()
), "Duplicate method string keys detected!"


class TestMode(IntEnum):
    """
    Types of modes that determine how the number of Voigt elements (capacitor connected in parallel with resistor) is chosen:

    - AUTO: follow procedure described by Schönleber, Klotz, and Ivers-Tiffée (2014)
    - EXPLORATORY: same procedure as AUTO but present intermediate results to user and apply additional weighting to the initial suggestion
    - MANUAL: manually choose the number
    """

    AUTO = 1
    EXPLORATORY = 2
    MANUAL = 3


test_mode_to_label: Dict[TestMode, str] = {
    TestMode.AUTO: "Auto",
    TestMode.EXPLORATORY: "Exploratory",
    TestMode.MANUAL: "Manual",
}
label_to_test_mode: Dict[str, TestMode] = {v: k for k, v in test_mode_to_label.items()}
assert set(test_mode_to_label.keys()) == set(
    label_to_test_mode.values()
), "Duplicate test mode string labels detected!"


class FitSimOutput(IntEnum):
    CDC_BASIC = auto()
    CDC_EXTENDED = auto()
    CSV_DATA_TABLE = auto()
    CSV_PARAMETERS_TABLE = auto()
    JSON_PARAMETERS_TABLE = auto()
    LATEX_DIAGRAM = auto()
    LATEX_EXPR = auto()
    LATEX_PARAMETERS_TABLE = auto()
    MARKDOWN_PARAMETERS_TABLE = auto()
    SVG_DIAGRAM = auto()
    SVG_DIAGRAM_NO_TERMINAL_LABELS = auto()
    SVG_DIAGRAM_NO_LABELS = auto()
    SYMPY_EXPR = auto()
    SYMPY_EXPR_VALUES = auto()


fit_sim_output_to_label: Dict[FitSimOutput, str] = {
    FitSimOutput.CDC_BASIC: "CDC - basic",
    FitSimOutput.CDC_EXTENDED: "CDC - extended",
    FitSimOutput.CSV_DATA_TABLE: "CSV - impedance table",
    FitSimOutput.CSV_PARAMETERS_TABLE: "CSV - parameters table",
    FitSimOutput.JSON_PARAMETERS_TABLE: "JSON - parameters table",
    FitSimOutput.LATEX_DIAGRAM: "LaTeX - circuit diagram",
    FitSimOutput.LATEX_EXPR: "LaTeX - expression",
    FitSimOutput.LATEX_PARAMETERS_TABLE: "LaTeX - parameters table",
    FitSimOutput.MARKDOWN_PARAMETERS_TABLE: "Markdown - parameters table",
    FitSimOutput.SVG_DIAGRAM: "SVG - circuit diagram",
    FitSimOutput.SVG_DIAGRAM_NO_TERMINAL_LABELS: "SVG - circuit diagram without terminal labels",
    FitSimOutput.SVG_DIAGRAM_NO_LABELS: "SVG - circuit diagram without any labels",
    FitSimOutput.SYMPY_EXPR: "SymPy - expression",
    FitSimOutput.SYMPY_EXPR_VALUES: "SymPy - expression and values",
}
label_to_fit_sim_output: Dict[str, FitSimOutput] = {
    v: k for k, v in fit_sim_output_to_label.items()
}
assert set(fit_sim_output_to_label.keys()) == set(
    label_to_fit_sim_output.values()
), "Duplicate output string labels detected!"


class PlotType(IntEnum):
    """
    Types of plots:

    - NYQUIST: -Zim vs Zre
    - BODE_MAGNITUDE: |Z| vs f
    - BODE_PHASE: phi vs f
    - DRT: gamma vs tau
    - IMPEDANCE_REAL: Zre vs f
    - IMPEDANCE_IMAGINARY: Zim vs f
    """

    NYQUIST = 1
    BODE_MAGNITUDE = 2
    BODE_PHASE = 3
    IMPEDANCE_REAL = 4
    IMPEDANCE_IMAGINARY = 5
    DRT = 6


plot_type_to_label: Dict[PlotType, str] = {
    PlotType.NYQUIST: "Nyquist",
    PlotType.BODE_MAGNITUDE: "Bode - magnitude",
    PlotType.BODE_PHASE: "Bode - phase",
    PlotType.IMPEDANCE_REAL: "Impedance - real",
    PlotType.IMPEDANCE_IMAGINARY: "Impedance - imaginary",
    PlotType.DRT: "Distribution of relaxation times",
}
label_to_plot_type: Dict[str, PlotType] = {v: k for k, v in plot_type_to_label.items()}
assert set(plot_type_to_label.keys()) == set(
    label_to_plot_type.values()
), "Duplicate plot type string labels detected!"


class Test(IntEnum):
    """
    Types of Kramers-Kronig tests:

    - CNLS: complex non-linear least-squares fit of circuit (fig. 1, Boukamp, 1995) with a distribution of fixed time constants
    - COMPLEX: eqs. 11 and 12, Boukamp, 1995
    - IMAGINARY: eqs. 4, 6, and 7, Boukamp, 1995
    - REAL: eqs. 5, 8, 9, and 10, Boukamp, 1995
    """

    CNLS = 1
    COMPLEX = 2
    IMAGINARY = 3
    REAL = 4


test_to_label: Dict[Test, str] = {
    Test.CNLS: "CNLS",
    Test.COMPLEX: "Complex",
    Test.IMAGINARY: "Imaginary",
    Test.REAL: "Real",
}
label_to_test: Dict[str, Test] = {v: k for k, v in test_to_label.items()}
test_to_value: Dict[Test, str] = {
    Test.CNLS: "cnls",
    Test.COMPLEX: "complex",
    Test.IMAGINARY: "imaginary",
    Test.REAL: "real",
}
assert set(test_to_label.keys()) == set(
    label_to_test.values()
), "Duplicate test string keys detected!"
assert set(test_to_label.keys()) == set(
    test_to_value.keys()
), "Missing test string values detected!"


class Weight(IntEnum):
    """
    Types of weights to use during complex non-linear least squares fitting:

    - AUTO: try each weight
    - BOUKAMP: 1 / (Zre^2 + Zim^2) (eq. 13, Boukamp, 1995)
    - MODULUS: 1 / |Z|
    - PROPORTIONAL: 1 / Zre^2, 1 / Zim^2
    - UNITY: 1
    """

    AUTO = 1
    UNITY = 2
    PROPORTIONAL = 3
    MODULUS = 4
    BOUKAMP = 5


weight_to_label: Dict[Weight, str] = {
    Weight.AUTO: "Auto",
    Weight.MODULUS: "Modulus",
    Weight.PROPORTIONAL: "Proportional",
    Weight.UNITY: "Unity",
    Weight.BOUKAMP: "Boukamp",
}
label_to_weight: Dict[str, Weight] = {v: k for k, v in weight_to_label.items()}
weight_to_value: Dict[Weight, str] = {
    Weight.AUTO: "auto",
    Weight.UNITY: "unity",
    Weight.PROPORTIONAL: "proportional",
    Weight.MODULUS: "modulus",
    Weight.BOUKAMP: "boukamp",
}
value_to_weight: Dict[str, Weight] = {v: k for k, v in weight_to_value.items()}
assert set(weight_to_label.keys()) == set(
    label_to_weight.values()
), "Duplicate weight string labels detected!"
assert set(weight_to_label.keys()) == set(
    weight_to_value.keys()
), "Missing weight string values detected!"
assert set(weight_to_value.keys()) == set(
    value_to_weight.values()
), "Duplicate weight string values detected!"


class DRTMethod(IntEnum):
    """
    The method to use for calculating the DRT:

    - TR_NNLS
    - TR_RBF
    - BHT
    - M_RQ_FIT
    """
    TR_NNLS = 1
    TR_RBF = 2
    BHT = 3
    M_RQ_FIT = 4


drt_method_to_label: Dict[DRTMethod, str] = {
    DRTMethod.BHT: "BHT",
    DRTMethod.TR_NNLS: "TR-NNLS",
    DRTMethod.TR_RBF: "TR-RBF",
    DRTMethod.M_RQ_FIT: "m(RQ)fit",
}
label_to_drt_method: Dict[str, DRTMethod] = {
    v: k for k, v in drt_method_to_label.items()
}
drt_method_to_value: Dict[DRTMethod, str] = {
    DRTMethod.BHT: "bht",
    DRTMethod.M_RQ_FIT: "m(RQ)fit",
    DRTMethod.TR_NNLS: "tr-nnls",
    DRTMethod.TR_RBF: "tr-rbf",
}
assert set(drt_method_to_label.keys()) == set(
    label_to_drt_method.values()
), "Duplicate DRT method string labels detected!"
assert set(drt_method_to_label.keys()) == set(
    drt_method_to_value.keys()
), "Missing DRT method string values detected!"


class DRTMode(IntEnum):
    """
    The parts of the impedance data to use:

    - COMPLEX
    - REAL
    - IMAGINARY
    """
    COMPLEX = 1
    REAL = 2
    IMAGINARY = 3


drt_mode_to_label: Dict[DRTMode, str] = {
    DRTMode.COMPLEX: "Complex",
    DRTMode.REAL: "Real",
    DRTMode.IMAGINARY: "Imaginary",
}
label_to_drt_mode: Dict[str, DRTMode] = {v: k for k, v in drt_mode_to_label.items()}
drt_mode_to_value: Dict[DRTMode, str] = {
    DRTMode.COMPLEX: "complex",
    DRTMode.REAL: "real",
    DRTMode.IMAGINARY: "imaginary",
}
assert set(drt_mode_to_label.keys()) == set(
    label_to_drt_mode.values()
), "Duplicate DRT mode string labels detected!"
assert set(drt_mode_to_label.keys()) == set(
    drt_mode_to_value.keys()
), "Missing DRT mode string values detected!"


class RBFType(IntEnum):
    """
    The radial basis function to use for discretization (or piecewise linear discretization):

    - C0_MATERN
    - C2_MATERN
    - C4_MATERN
    - C6_MATERN
    - CAUCHY
    - GAUSSIAN
    - INVERSE_QUADRATIC
    - INVERSE_QUADRIC
    - PIECEWISE_LINEAR
    """
    C0_MATERN = 1
    C2_MATERN = 2
    C4_MATERN = 3
    C6_MATERN = 4
    CAUCHY = 5
    GAUSSIAN = 6
    INVERSE_QUADRATIC = 7
    INVERSE_QUADRIC = 8
    PIECEWISE_LINEAR = 9


rbf_type_to_label: Dict[RBFType, str] = {
    RBFType.C0_MATERN: "C^0 Matérn",
    RBFType.C2_MATERN: "C^2 Matérn",
    RBFType.C4_MATERN: "C^4 Matérn",
    RBFType.C6_MATERN: "C^6 Matérn",
    RBFType.CAUCHY: "Cauchy",
    RBFType.GAUSSIAN: "Gaussian",
    RBFType.INVERSE_QUADRATIC: "Inverse quadratic",
    RBFType.INVERSE_QUADRIC: "Inverse quadric",
    RBFType.PIECEWISE_LINEAR: "Piecewise linear",
}
label_to_rbf_type: Dict[str, RBFType] = {v: k for k, v in rbf_type_to_label.items()}
rbf_type_to_value: Dict[RBFType, str] = {
    RBFType.C0_MATERN: "c0-matern",
    RBFType.C2_MATERN: "c2-matern",
    RBFType.C4_MATERN: "c4-matern",
    RBFType.C6_MATERN: "c6-matern",
    RBFType.CAUCHY: "cauchy",
    RBFType.GAUSSIAN: "gaussian",
    RBFType.INVERSE_QUADRATIC: "inverse-quadratic",
    RBFType.INVERSE_QUADRIC: "inverse-quadric",
    RBFType.PIECEWISE_LINEAR: "piecewise-linear",
}
assert set(rbf_type_to_label.keys()) == set(
    label_to_rbf_type.values()
), "Duplicate RBF type string labels detected!"
assert set(rbf_type_to_label.keys()) == set(
    rbf_type_to_value.keys()
), "Missing RBF type string values detected!"


class RBFShape(IntEnum):
    """
    The shape to use with the radial basis function discretization:

    - FWHM
    - FACTOR
    """
    FWHM = 1
    FACTOR = 2


rbf_shape_to_label: Dict[RBFShape, str] = {
    RBFShape.FACTOR: "Factor",
    RBFShape.FWHM: "FWHM",
}
label_to_rbf_shape: Dict[str, RBFShape] = {v: k for k, v in rbf_shape_to_label.items()}
rbf_shape_to_value: Dict[RBFShape, str] = {
    RBFShape.FACTOR: "factor",
    RBFShape.FWHM: "fwhm",
}
assert set(rbf_shape_to_label.keys()) == set(
    label_to_rbf_shape.values()
), "Duplicate RBF shape string labels detected!"
assert set(rbf_shape_to_label.keys()) == set(
    rbf_shape_to_value.keys()
), "Missing RBF shape string values detected!"


derivative_order_to_label: Dict[int, str] = {
    1: "1st",
    2: "2nd",
}
label_to_derivative_order: Dict[str, int] = {
    v: k for k, v in derivative_order_to_label.items()
}


class DRTOutput(IntEnum):
    CSV_SCORES = auto()
    JSON_SCORES = auto()
    LATEX_SCORES = auto()
    MARKDOWN_SCORES = auto()


drt_output_to_label: Dict[DRTOutput, str] = {
    DRTOutput.CSV_SCORES: "CSV - scores",
    DRTOutput.JSON_SCORES: "JSON - scores",
    DRTOutput.LATEX_SCORES: "LaTeX - scores",
    DRTOutput.MARKDOWN_SCORES: "Markdown - scores",
}
label_to_drt_output: Dict[str, DRTOutput] = {
    v: k for k, v in drt_output_to_label.items()
}
assert set(drt_output_to_label.keys()) == set(
    label_to_drt_output.values()
), "Duplicate DRT output string labels detected!"


class PlotUnits(IntEnum):
    """
    The units of the plot dimensions:
    
    - INCHES
    - CENTIMETERS
    """
    INCHES = 1
    CENTIMETERS = 2


plot_units_to_label: Dict[PlotUnits, str] = {
    PlotUnits.INCHES: "Inches",
    PlotUnits.CENTIMETERS: "Centimeters",
}
label_to_plot_units: Dict[str, PlotUnits] = {
    v: k for k, v in plot_units_to_label.items()
}
plot_units_per_inch: Dict[PlotUnits, float] = {
    PlotUnits.INCHES: 1.0,
    PlotUnits.CENTIMETERS: 2.54,
}
assert set(plot_units_to_label.keys()) == set(
    label_to_plot_units.values()
), "Duplicate plot unit string labels detected!"
assert set(plot_units_to_label.keys()) == set(
    plot_units_per_inch.keys()
), "Missing plot unit string values detected!"


class PlotPreviewLimit(IntEnum):
    """
    The limits of the plot preview:

    - NONE
    - PX256
    - PX512
    - PX1024
    - PX2048
    - PX4096
    - PX8192
    - PX16384
    """
    NONE = 0
    PX256 = 8
    PX512 = 9
    PX1024 = 10
    PX2048 = 11
    PX4096 = 12
    PX8192 = 13
    PX16384 = 14


plot_preview_limit_to_label: Dict[PlotPreviewLimit, str] = {
    PlotPreviewLimit.NONE: "No limit",
    PlotPreviewLimit.PX256: f"{2**8} px",
    PlotPreviewLimit.PX512: f"{2**9} px",
    PlotPreviewLimit.PX1024: f"{2**10} px",
    PlotPreviewLimit.PX2048: f"{2**11} px",
    PlotPreviewLimit.PX4096: f"{2**12} px",
    PlotPreviewLimit.PX8192: f"{2**13} px",
    PlotPreviewLimit.PX16384: f"{2**14} px",
}
label_to_plot_preview_limit: Dict[str, PlotPreviewLimit] = {
    v: k for k, v in plot_preview_limit_to_label.items()
}
assert set(plot_preview_limit_to_label.keys()) == set(
    label_to_plot_preview_limit.values()
), "Duplicate plot preview limit string labels detected!"


class PlotLegendLocation(IntEnum):
    """
    The position of the plot legend:

    - AUTO
    - UPPER_RIGHT
    - UPPER_LEFT
    - LOWER_LEFT
    - LOWER_RIGHT
    - RIGHT
    - CENTER_LEFT
    - CENTER_RIGHT
    - LOWER_CENTER
    - UPPER_CENTER
    - CENTER
    """
    AUTO = 0
    UPPER_RIGHT = 1
    UPPER_LEFT = 2
    LOWER_LEFT = 3
    LOWER_RIGHT = 4
    RIGHT = 5
    CENTER_LEFT = 6
    CENTER_RIGHT = 7
    LOWER_CENTER = 8
    UPPER_CENTER = 9
    CENTER = 10


plot_legend_location_to_label: Dict[PlotLegendLocation, str] = {
    PlotLegendLocation.AUTO: "Automatic",
    PlotLegendLocation.UPPER_RIGHT: "Upper right",
    PlotLegendLocation.UPPER_LEFT: "Upper left",
    PlotLegendLocation.LOWER_LEFT: "Lower left",
    PlotLegendLocation.LOWER_RIGHT: "Lower right",
    PlotLegendLocation.RIGHT: "Right",
    PlotLegendLocation.CENTER_LEFT: "Center left",
    PlotLegendLocation.CENTER_RIGHT: "Center right",
    PlotLegendLocation.LOWER_CENTER: "Lower center",
    PlotLegendLocation.UPPER_CENTER: "Upper center",
    PlotLegendLocation.CENTER: "Center",
}
label_to_plot_legend_location: Dict[str, PlotLegendLocation] = {
    v: k for k, v in plot_legend_location_to_label.items()
}
assert set(plot_legend_location_to_label.keys()) == set(
    label_to_plot_legend_location.values()
), "Duplicate plot legend location string labels detected!"


PLOT_EXTENSIONS: List[str] = [
    ".eps",
    ".jpg",
    ".pdf",
    ".pgf",
    ".png",
    ".ps",
    ".svg",
]



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from time import time as _time
from typing import (
    Dict,
    List,
)
from uuid import uuid4 as _uuid4
from numpy import (
    integer as _integer,
    issubdtype as _issubdtype,
)
import pyimpspec as _pyimpspec
from deareis.data import (
    DataSet,
    TestResult,
    TestSettings,
)
from deareis.enums import (
    CNLSMethod,
    TestMode,
    Test,
    cnls_method_to_value as _cnls_method_to_value,
    test_to_value as _test_to_value,
)


def perform_test(
    data: DataSet,
    settings: TestSettings,
    num_procs: int = -1,
) -> TestResult:
    """
    Wrapper for the `pyimpspec.perform_test` function.

    Performs a linear Kramers-Kronig test as described by Boukamp (1995).
    The results can be used to check the validity of an impedance spectrum before performing equivalent circuit fitting.
    If the number of (RC) circuits is less than two, then a suitable number of (RC) circuits is determined using the procedure described by Schönleber et al. (2014) based on a criterion for the calculated mu-value (zero to one).
    A mu-value of one represents underfitting and a mu-value of zero represents overfitting.

    References:

    - B.A. Boukamp, 1995, J. Electrochem. Soc., 142, 1885-1894 (https://doi.org/10.1149/1.2044210)
    - M. Schönleber, D. Klotz, and E. Ivers-Tiffée, 2014, Electrochim. Acta, 131, 20-27 (https://doi.org/10.1016/j.electacta.2014.01.034)

    Parameters
    ----------
    data: DataSet
        The data to be tested.

    settings: TestSettings
        The settings that determine how the test is performed.
        Note that `Test.EXPLORATORY` is not supported by this function.

    num_procs: int = -1
        The maximum number of parallel processes to use when performing a test.
        A value less than one results in using the number of cores returned by multiprocessing.cpu_count.
        Applies only to the `TestMode.CNLS` test.

    Returns
    -------
    TestResult
    """
    assert isinstance(data, _pyimpspec.DataSet), data
    assert type(settings) is TestSettings, settings
    assert (
        settings.mode != TestMode.EXPLORATORY
    ), "Use deareis.perform_exploratory_tests to perform the tests!"
    assert _issubdtype(type(num_procs), _integer), num_procs
    result: _pyimpspec.TestResult = _pyimpspec.perform_test(
        data=data,
        test=_test_to_value[settings.test],
        num_RC=settings.num_RC * (-1 if settings.mode == TestMode.AUTO else 1),
        mu_criterion=settings.mu_criterion,
        add_capacitance=settings.add_capacitance,
        add_inductance=settings.add_inductance,
        method=_cnls_method_to_value[settings.method],
        max_nfev=settings.max_nfev,
        num_procs=num_procs,
    )
    return TestResult(
        _uuid4().hex,
        _time(),
        result.circuit,
        result.num_RC,
        result.mu,
        result.pseudo_chisqr,
        result.frequency,
        result.impedance,
        result.real_residual,
        result.imaginary_residual,
        data.get_mask().copy(),
        settings,
    )


def perform_exploratory_tests(
    data: DataSet,
    settings: TestSettings,
    num_procs: int = -1,
) -> List[TestResult]:
    """
    Wrapper for the `pyimpspec.perform_exploratory_tests` function.

    Performs a batch of linear Kramers-Kronig tests, which are then scored and sorted from best to worst before they are returned.

    Parameters
    ----------
    data: DataSet
        The data set to be tested.

    settings: TestSettings
        The settings that determine how the test is performed.
        Note that only `Test.EXPLORATORY` is supported by this function.

    num_procs: int = -1
        See perform_test for details.

    Returns
    -------
    List[TestResult]
    """
    assert (
        settings.mode == TestMode.EXPLORATORY
    ), "Use deareis.perform_test to perform the test!"
    num_RCs: List[int] = list(range(1, settings.num_RC + 1))
    assert len(num_RCs) > 0, "Invalid settings!"
    results: List[_pyimpspec.TestResult] = _pyimpspec.perform_exploratory_tests(
        data=data,
        test=_test_to_value[settings.test],
        num_RCs=num_RCs,
        mu_criterion=settings.mu_criterion,
        add_capacitance=settings.add_capacitance,
        add_inductance=settings.add_inductance,
        method=_cnls_method_to_value[settings.method],
        max_nfev=settings.max_nfev,
        num_procs=num_procs,
    )
    time: float = _time()
    mask: Dict[int, bool] = data.get_mask()
    return list(
        map(
            lambda _: TestResult(
                _uuid4().hex,
                time,
                _.circuit,
                _.num_RC,
                _.mu,
                _.pseudo_chisqr,
                _.frequency,
                _.impedance,
                _.real_residual,
                _.imaginary_residual,
                mask.copy(),
                settings,
            ),
            results,
        )
    )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace


simulation = SimpleNamespace(
    **{
        "cdc": """
The equivalent circuit to simulate. A circuit description code (CDC) can be typed in to define the equivalent circuit. Alternatively, the circuit editor can be used to construct the equivalent circuit.
    """.strip(),
        "min_freq": """
The minimum frequency to simulate.
    """.strip(),
        "max_freq": """
The maximum frequency to simulate.
    """.strip(),
        "per_decade": """
The number of frequencies per decade in the frequency range.
    """.strip(),
        "element": """
An element in the circuit.
    """.strip(),
        "parameter": """
A parameter in the element.
    """.strip(),
        "parameter_value": """
The value of the parameter.
    """.strip(),
        "remove": """
Remove the current simulation result.
    """.strip(),
        "load_as_data_set": """
Load the current simulation as a data set.
    """.strip(),
    },
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Optional,
)
from deareis.data import (
    DataSet,
    PlotSettings,
    Project,
    SimulationResult,
    SimulationSettings,
)
import deareis.api.simulation as api
from deareis.enums import Context
from deareis.gui import ProjectTab
from deareis.signals import Signal
from deareis.state import STATE
import deareis.signals as signals


def select_simulation_result(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    simulation: Optional[SimulationResult] = kwargs.get("simulation")
    data: Optional[DataSet] = kwargs.get("data")
    is_busy_message_visible: bool = STATE.is_busy_message_visible()
    if not is_busy_message_visible:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Loading simulation result")
    project_tab.select_simulation_result(simulation, data)
    if not is_busy_message_visible:
        signals.emit(Signal.HIDE_BUSY_MESSAGE)


def delete_simulation_result(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    simulation: Optional[SimulationResult] = kwargs.get("simulation")
    if simulation is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Deleting simulation result")
    settings: Optional[PlotSettings] = project_tab.get_active_plot()
    update_plot: bool = (
        simulation.uuid in settings.series_order if settings is not None else False
    )
    project.delete_simulation(simulation)
    project_tab.populate_simulations(project)
    if settings is not None:
        project_tab.plotting_tab.populate_simulations(
            project.get_simulations(),
            settings,
        )
    if update_plot:
        signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def apply_simulation_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[SimulationSettings] = kwargs.get("settings")
    if settings is None:
        return
    project_tab.set_simulation_settings(settings)


def perform_simulation(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[SimulationSettings] = kwargs.get("settings")
    if settings is None:
        return
    assert (
        settings.min_frequency != settings.max_frequency
    ), "The minimum and maximum frequencies cannot be the same!"
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Performing simulation")
    simulation: SimulationResult = api.simulate_spectrum(settings)
    project.add_simulation(simulation)
    project_tab.populate_simulations(project)
    signals.emit(
        Signal.SELECT_SIMULATION_RESULT,
        simulation=simulation,
        data=project_tab.get_active_data_set(context=Context.SIMULATION_TAB),
    )
    project_tab.plotting_tab.populate_simulations(
        project.get_simulations(),
        project_tab.get_active_plot(),
    )
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from traceback import format_exc
from enum import (
    IntEnum,
    auto,
)
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)


DEBUG: bool = False


class Signal(IntEnum):
    APPLY_DATA_SET_MASK = auto()
    APPLY_DRT_SETTINGS = auto()
    APPLY_FIT_SETTINGS = auto()
    APPLY_SIMULATION_SETTINGS = auto()
    APPLY_TEST_SETTINGS = auto()
    AVERAGE_DATA_SETS = auto()
    BLOCK_KEYBINDINGS = auto()
    CHECK_UPDATES = auto()
    CLEAR_RECENT_PROJECTS = auto()
    CLOSE_PROJECT = auto()
    COPY_OUTPUT = auto()
    COPY_PLOT_APPEARANCE_SETTINGS = auto()
    COPY_PLOT_DATA = auto()
    CREATE_PROJECT_SNAPSHOT = auto()
    DELETE_DATA_SET = auto()
    DELETE_DRT_RESULT = auto()
    DELETE_FIT_RESULT = auto()
    DELETE_PLOT_SETTINGS = auto()
    DELETE_SIMULATION_RESULT = auto()
    DELETE_TEST_RESULT = auto()
    EXPORT_PLOT = auto()
    HIDE_BUSY_MESSAGE = auto()
    LOAD_DATA_SET_FILES = auto()
    LOAD_PROJECT_FILES = auto()
    LOAD_SIMULATION_AS_DATA_SET = auto()
    MODIFY_DATA_SET_PATH = auto()
    MODIFY_PLOT_SERIES_THEME = auto()
    MODIFY_PROJECT_NOTES = auto()
    NEW_PLOT_SETTINGS = auto()
    NEW_PROJECT = auto()
    PERFORM_DRT = auto()
    PERFORM_FIT = auto()
    PERFORM_SIMULATION = auto()
    PERFORM_TEST = auto()
    REDO_PROJECT_ACTION = auto()
    RENAME_DATA_SET = auto()
    RENAME_PLOT_SERIES = auto()
    RENAME_PLOT_SETTINGS = auto()
    RENAME_PROJECT = auto()
    RENDER_MATH = auto()
    REORDER_PLOT_SERIES = auto()
    RESTORE_PROJECT_STATE = auto()
    SAVE_PLOT = auto()
    SAVE_PROJECT = auto()
    SAVE_PROJECT_AS = auto()
    SELECT_DATA_POINTS_TO_TOGGLE = auto()
    SELECT_DATA_SET = auto()
    SELECT_DATA_SETS_TO_AVERAGE = auto()
    SELECT_DATA_SET_FILES = auto()
    SELECT_DATA_SET_MASK_TO_COPY = auto()
    SELECT_DRT_RESULT = auto()
    SELECT_FIT_RESULT = auto()
    SELECT_HOME_TAB = auto()
    SELECT_IMPEDANCE_TO_SUBTRACT = auto()
    SELECT_PLOT_APPEARANCE_SETTINGS = auto()
    SELECT_PLOT_SETTINGS = auto()
    SELECT_PLOT_TYPE = auto()
    SELECT_PROJECT_FILES = auto()
    SELECT_PROJECT_TAB = auto()
    SELECT_SIMULATION_RESULT = auto()
    SELECT_TEST_RESULT = auto()
    SHOW_BUSY_MESSAGE = auto()
    SHOW_CHANGELOG = auto()
    SHOW_COMMAND_PALETTE = auto()
    SHOW_ENLARGED_PLOT = auto()
    SHOW_ERROR_MESSAGE = auto()
    SHOW_HELP_ABOUT = auto()
    SHOW_HELP_LICENSES = auto()
    SHOW_SETTINGS_APPEARANCE = auto()
    SHOW_SETTINGS_DEFAULTS = auto()
    SHOW_SETTINGS_KEYBINDINGS = auto()
    TOGGLE_DATA_POINT = auto()
    TOGGLE_PLOT_SERIES = auto()
    UNBLOCK_KEYBINDINGS = auto()
    UNDO_PROJECT_ACTION = auto()
    VIEWPORT_RESIZED = auto()


_UUID_COUNTER: int = 0
_REGISTERED_CALLBACKS: Dict[Signal, List[Tuple[Callable, int]]] = {}
_QUEUE: Optional[Dict[Signal, List[Tuple[tuple, dict]]]] = {}


def emit(signal: Signal, *args, **kwargs):
    global _REGISTERED_CALLBACKS
    global _QUEUE
    assert type(signal) is Signal, signal
    if _QUEUE is not None and signal not in _REGISTERED_CALLBACKS:
        if signal not in _QUEUE:
            _QUEUE[signal] = []
        _QUEUE[signal].append(
            (
                args,
                kwargs,
            )
        )
        return
    assert signal in _REGISTERED_CALLBACKS, signal
    try:
        if DEBUG:
            print(f"\nsignals.emit: {str(signal)}")
            if args:
                print("- args:")
                arg: Any
                for arg in args:
                    if len(repr(arg)) > 80:
                        print(f" - {repr(arg)[:80]}...")
                    else:
                        print(f" - {repr(arg)}")
            if kwargs:
                print("- kwargs:")
                key: str
                value: Any
                for key, value in kwargs.items():
                    if len(repr(value)) > 80:
                        print(f" - {key}: {repr(value)[:80]}...")
                    else:
                        print(f" - {key}: {repr(value)}")
            print("- entries:")
        entry: Tuple[Callable, int]
        func: Callable
        uuid: int
        for (func, uuid) in _REGISTERED_CALLBACKS[signal]:
            if DEBUG:
                print(f" - {uuid}: {repr(func)}")
            func(*args, **kwargs)
    except Exception:
        if signal == Signal.SHOW_ERROR_MESSAGE:
            print(format_exc())
        else:
            emit(Signal.SHOW_ERROR_MESSAGE, format_exc())


def register(signal: Signal, callback: Callable) -> int:
    global _UUID_COUNTER
    global _REGISTERED_CALLBACKS
    assert type(signal) is Signal, signal
    try:
        if signal not in _REGISTERED_CALLBACKS:
            _REGISTERED_CALLBACKS[signal] = []
        _UUID_COUNTER += 1
        _REGISTERED_CALLBACKS[signal].append(
            (
                callback,
                _UUID_COUNTER,
            )
        )
        if DEBUG:
            print(f"\nsignals.register: {str(signal)}")
            print(f"- callback: {callback} ({_UUID_COUNTER})")
        return _UUID_COUNTER
    except Exception:
        if signal == Signal.SHOW_ERROR_MESSAGE:
            print(format_exc())
        else:
            emit(Signal.SHOW_ERROR_MESSAGE, format_exc())
        return -1


def unregister(
    signal: Signal, callback: Optional[Callable] = None, uuid: Optional[int] = None
):
    if DEBUG:
        print(f"\nsignals.unregister: {str(signal)}")
        print(f"- callback: {callback}")
        print(f"- uuid: {uuid}")
    global _REGISTERED_CALLBACKS
    assert type(signal) is Signal, signal
    assert callback is not None or type(uuid) is int and uuid > 0, (
        callback,
        uuid,
    )
    assert signal in _REGISTERED_CALLBACKS, signal
    try:
        chosen_entry: Optional[Tuple[Callable, int]] = None
        entry: Tuple[Callable, int]
        for entry in _REGISTERED_CALLBACKS[signal]:
            if entry[0] == callback or entry[1] == uuid:
                chosen_entry = entry
                break
        assert chosen_entry is not None
        _REGISTERED_CALLBACKS[signal].remove(chosen_entry)
    except Exception:
        if signal == Signal.SHOW_ERROR_MESSAGE:
            print(format_exc())
        else:
            emit(Signal.SHOW_ERROR_MESSAGE, format_exc())


def clear(signal: Signal):
    if DEBUG:
        print(f"\nsignals.clear: {str(signal)}")
    global _REGISTERED_CALLBACKS
    assert type(signal) is Signal, signal
    assert signal in _REGISTERED_CALLBACKS, signal
    try:
        del _REGISTERED_CALLBACKS[signal]
    except Exception:
        if signal == Signal.SHOW_ERROR_MESSAGE:
            print(format_exc())
        else:
            emit(Signal.SHOW_ERROR_MESSAGE, format_exc())


def emit_backlog():
    global _QUEUE
    signal: Signal
    queue: List[Tuple[list, dict]]
    for signal, queue in _QUEUE.items():
        for (args, kwargs) in queue:
            emit(signal, *args, **kwargs)
    _QUEUE = None



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from os.path import (
    abspath,
    dirname,
    exists,
    join,
)
from typing import (
    IO,
    List,
)
from deareis.signals import Signal
from deareis.utility import calculate_window_position_dimensions
import deareis.signals as signals
import dearpygui.dearpygui as dpg


def format_changelog(changelog: str, width: int) -> str:
    lines: List[str] = []
    line: str
    for line in changelog.split("\n"):
        stripped_line: str = line.strip()
        if stripped_line.startswith("#"):
            lines.append(line)
            continue
        elif stripped_line == "":
            lines.append("")
            continue
        indentation: str = ""
        if line.lstrip() != line:
            indentation = line[: line.index(stripped_line)]
            indentation = indentation.replace("\t", "  ")
        fragments: List[str] = [indentation] if indentation != "" else []
        if stripped_line.startswith("- "):
            indentation += "  "
        words: List[str] = stripped_line.split(" ")
        while words:
            while words:
                fragments.append(words.pop(0))
                if dpg.get_text_size(" ".join(fragments), wrap_width=-1.0)[0] >= width:
                    words.insert(0, fragments.pop())
                    break
            lines.append(" ".join(fragments))
            fragments = [indentation] if indentation != "" else []
        if len(fragments) > 1:
            lines.append(" ".join(fragments))
    i: int
    for i in range(0, len(lines)):
        line = lines[i]
        if line == "":
            continue
        indentation = line[: line.index(line.strip())]
        while len(indentation) % 2 > 0:
            indentation = indentation[1:]
        lines[i] = indentation + line.strip()
    return "\n".join(lines)


def show_changelog():
    changelog_path: str = join(dirname(abspath(__file__)), "CHANGELOG.md")
    assert exists(changelog_path), changelog_path
    changelog: str = ""
    fp: IO
    with open(changelog_path, "r") as fp:
        changelog = fp.read()

    window: int = dpg.generate_uuid()
    key_handler: int = dpg.generate_uuid()

    def close_window():
        if dpg.does_item_exist(window):
            dpg.delete_item(window)
        if dpg.does_item_exist(key_handler):
            dpg.delete_item(key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    with dpg.handler_registry(tag=key_handler):
        dpg.add_key_release_handler(
            key=dpg.mvKey_Escape,
            callback=close_window,
        )

    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions(640, 540)
    with dpg.window(
        label="Changelog",
        modal=True,
        no_resize=True,
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        on_close=close_window,
        tag=window,
    ):
        dpg.add_text(format_changelog(changelog, w - 40))
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window, window_object=None)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg


class BusyMessage:
    def __init__(self):
        self.width: int = 159
        self.height: int = 159
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            autosize=True,
            menubar=False,
            modal=True,
            no_background=True,
            no_close=True,
            no_move=True,
            no_resize=True,
            no_title_bar=True,
            show=False,
            tag=self.window,
        ):
            dpg.add_loading_indicator(
                radius=11,
                style=0,
                speed=2,
                color=(
                    66,
                    139,
                    255,
                    190,
                ),
            )
            with dpg.group(horizontal=True):
                self.message_spacer: int = dpg.generate_uuid()
                dpg.add_spacer(tag=self.message_spacer)
                self.message_wrap: int = 140
                self.message_text: int = dpg.generate_uuid()
                dpg.add_text(tag=self.message_text, wrap=self.message_wrap)
            self.progress_bar: int = dpg.generate_uuid()
            dpg.add_progress_bar(
                width=self.width - 16,
                height=12,
                tag=self.progress_bar,
            )

    def is_visible(self) -> bool:
        return dpg.is_item_shown(self.window)

    def show(self, message: str = "", progress: float = -1.0):
        assert type(message) is str, message
        assert type(progress) is float and progress <= 1.0, progress
        dpg.split_frame(delay=33)
        if not self.is_visible():
            dpg.split_frame()
            dpg.show_item(self.window)
        if message == "":
            dpg.hide_item(self.message_text)
        else:
            dpg.show_item(self.message_text)
            dpg.set_item_width(
                self.message_spacer,
                max(
                    0,
                    (
                        self.width
                        - dpg.get_text_size(message, wrap_width=self.message_wrap)[0]
                    )
                    / 2
                    - 16,
                ),
            )
            dpg.set_value(self.message_text, message)
        if progress < 0.0:
            dpg.hide_item(self.progress_bar)
        else:
            dpg.show_item(self.progress_bar)
            dpg.configure_item(
                self.progress_bar,
                default_value=progress,
            )

    def hide(self):
        dpg.hide_item(self.window)
        dpg.split_frame()

    def resize(self, width: int, height: int):
        assert type(width) is int
        assert type(height) is int
        x: int = round((width - self.width) / 2)
        y: int = round((height - self.height) / 2)
        dpg.configure_item(
            self.window,
            pos=(
                x,
                y,
            ),
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    Optional,
)
from numpy import allclose
import pyimpspec


class DataSet(pyimpspec.DataSet):
    """
    Extends `pyimpspec.DataSet` to implement data minimization when writing to disk and to recreate the data when loading from disk.
    Equality checks between DataSet instances is also modified.

    Parameters
    ----------
    frequency: ndarray
        A 1-dimensional array of frequencies in hertz.

    impedance: ndarray
        A 1-dimensional array of complex impedances in ohms.

    mask: Dict[int, bool] = {}
        A mapping of integer indices to boolean values where a value of True means that the data point is to be omitted.

    path: str = ""
        The path to the file that has been parsed to generate this DataSet instance.

    label: str = ""
        The label assigned to this DataSet instance.

    uuid: str = ""
        The universivally unique identifier assigned to this DataSet instance.
        If empty, then one will be automatically assigned.
    """

    def __eq__(self, other) -> bool:
        # This is implemented because gui/data_sets.py checks if the newly selected DataSet is the
        # same as the current DataSet (if there even is one) and then decides whether to clear the
        # table of data points or if it just needs to be updated. Clearing the table causes the
        # scroll position to reset, which results in a bad UX when the user is modifying the
        # DataSet's mask.
        try:
            assert isinstance(other, pyimpspec.DataSet), other
            assert self.uuid == other.uuid, (
                self.uuid,
                other.uuid,
            )
            assert self.get_label() == other.get_label(), (
                self.get_label(),
                other.get_label(),
            )
            assert self.get_path() == other.get_path(), (
                self.get_path(),
                other.get_path(),
            )
            assert self.get_num_points(masked=None) == other.get_num_points(
                masked=None
            ), (
                self.get_num_points(masked=None),
                other.get_num_points(masked=None),
            )
            assert allclose(
                self.get_frequency(masked=None), other.get_frequency(masked=None)
            )
            assert allclose(
                self.get_impedance(masked=None), other.get_impedance(masked=None)
            )
        except AssertionError:
            return False
        return True

    def to_dict(self, session: bool = True) -> dict:
        """
        Return a dictionary that can be used to recreate this data set.

        Parameters
        ----------
        session: bool = True
            If true, then no data minimization is performed.
        """
        # This is implemented to reduce the file sizes of projects.
        dictionary: dict = super().to_dict()
        if not session:
            dictionary["mask"] = {
                k: v for k, v in dictionary["mask"].items() if v is True
            }
        return dictionary

    @classmethod
    def from_dict(Class, dictionary: dict) -> "DataSet":
        """
        Return a DataSet instance that has been created based off of a dictionary.

        Parameters
        ----------
        dictionary: dict
            Create an instance from a dictionary.
        """
        # This is implemented to deal with the effects of the modified to_dict method.
        mask: Optional[Dict[str, bool]] = dictionary.get("mask")
        if type(mask) is dict and len(mask) < len(dictionary["frequency"]):
            i: int
            for i in range(0, len(dictionary["frequency"])):
                if mask.get(str(i)) is not True:
                    mask[str(i)] = False
        return Class(**Class._parse(dictionary))



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)
from numpy import (
    allclose,
    array,
    log10 as log,
    ndarray,
)
from pyimpspec import (
    Circuit,
    Element,
    FittedParameter,
)
import pyimpspec
from pyimpspec.analysis.fitting import _calculate_residuals
import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.enums import (
    Context,
    CNLSMethod,
    FitSimOutput,
    Weight,
    label_to_cnls_method,
    label_to_weight,
    label_to_fit_sim_output,
    cnls_method_to_label,
    weight_to_label,
)
from deareis.data import (
    DataSet,
    FitResult,
    FitSettings,
)
from deareis.gui.plots import (
    Bode,
    Nyquist,
    Residuals,
)
import deareis.tooltips as tooltips
from deareis.tooltips import (
    attach_tooltip,
    update_tooltip,
)
from deareis.utility import (
    align_numbers,
    calculate_window_position_dimensions,
    format_number,
    render_math,
)
import deareis.themes as themes
from deareis.gui.circuit_editor import (
    CircuitPreview,
    CircuitEditor,
)


MATH_WEIGHT_WIDTH: int = 300
MATH_WEIGHT_HEIGHT: int = 40
MATH_WEIGHT_MODULUS: int = -1
MATH_WEIGHT_PROPORTIONAL: int = -1
MATH_WEIGHT_UNITY: int = -1
MATH_WEIGHT_BOUKAMP: int = -1
MATH_Z_FIT: int = -1
MATH_Z_EXP: int = -1


def process_math():
    global MATH_WEIGHT_MODULUS
    global MATH_WEIGHT_PROPORTIONAL
    global MATH_WEIGHT_UNITY
    global MATH_WEIGHT_BOUKAMP
    global MATH_Z_FIT
    global MATH_Z_EXP
    MATH_WEIGHT_MODULUS = render_math(
        r"$w_i = [|Z_{\rm re}(f_i)|, |Z_{\rm im}(f_i)|]^{-1}$",
        width=MATH_WEIGHT_WIDTH,
        height=MATH_WEIGHT_HEIGHT,
    )
    MATH_WEIGHT_PROPORTIONAL = render_math(
        r"$w_i = [Z_{\rm re}(f_i)^2, Z_{\rm im}(f_i)^2]^{-1}$",
        width=MATH_WEIGHT_WIDTH,
        height=MATH_WEIGHT_HEIGHT,
    )
    MATH_WEIGHT_UNITY = render_math(
        r"$w_i = [1]$",
        width=MATH_WEIGHT_WIDTH,
        height=MATH_WEIGHT_HEIGHT,
    )
    MATH_WEIGHT_BOUKAMP = render_math(
        r"$w_i = [(Z_{{\rm re},i})^2 + (Z_{{\rm im},i})^2]^{-1}$",
        width=MATH_WEIGHT_WIDTH,
        height=MATH_WEIGHT_HEIGHT,
    )
    MATH_Z_FIT = render_math(
        r"$Z_{\rm re/im}(f_i)$",
        width=54,
        height=20,
        fontsize=10,
    )
    MATH_Z_EXP = render_math(
        r"$Z_{{\rm re/im},i}$",
        width=44,
        height=20,
        fontsize=10,
    )


signals.register(Signal.RENDER_MATH, process_math)


class SettingsMenu:
    def __init__(
        self,
        default_settings: FitSettings,
        label_pad: int,
        circuit_editor: Optional[CircuitEditor] = None,
    ):
        self.circuit_editor: Optional[CircuitEditor] = circuit_editor
        with dpg.group(horizontal=True):
            dpg.add_text("Circuit".rjust(label_pad))
            attach_tooltip(tooltips.fitting.cdc)
            self.cdc_input: int = dpg.generate_uuid()
            dpg.add_input_text(
                width=-50 if circuit_editor is not None else -1,
                tag=self.cdc_input,
                on_enter=True,
                callback=lambda s, a, u: self.parse_cdc(a, s),
            )
            self.cdc_tooltip: int = dpg.generate_uuid()
            attach_tooltip("", tag=self.cdc_tooltip, parent=self.cdc_input)
            dpg.hide_item(dpg.get_item_parent(self.cdc_tooltip))
            if circuit_editor is not None:
                self.editor_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Edit",
                    callback=self.show_circuit_editor,
                    width=-1,
                    tag=self.editor_button,
                )
                attach_tooltip(tooltips.general.open_circuit_editor)
        with dpg.group(horizontal=True):
            dpg.add_text("Method".rjust(label_pad))
            attach_tooltip(tooltips.fitting.method)
            self.method_combo: int = dpg.generate_uuid()
            dpg.add_combo(
                items=list(label_to_cnls_method.keys()),
                default_value="Auto",
                width=-1,
                tag=self.method_combo,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Weight".rjust(label_pad))
            with dpg.tooltip(dpg.last_item()):
                dpg.add_text(
                    "The weight function used when calculating residuals during fitting."
                )
                with dpg.table(
                    borders_outerV=True,
                    borders_outerH=True,
                    borders_innerV=True,
                    borders_innerH=True,
                    scrollY=True,
                    freeze_rows=1,
                    height=18 + 4 * (MATH_WEIGHT_HEIGHT + 4),
                ):
                    dpg.add_table_column(
                        label="Label",
                        width_fixed=True,
                    )
                    dpg.add_table_column(
                        label="Equation",
                        width_fixed=False,
                    )
                    with dpg.table_row():
                        dpg.add_text("Modulus")
                        dpg.add_image(MATH_WEIGHT_MODULUS)
                    with dpg.table_row():
                        dpg.add_text("Proportional")
                        dpg.add_image(MATH_WEIGHT_PROPORTIONAL)
                    with dpg.table_row():
                        dpg.add_text("Unity")
                        dpg.add_image(MATH_WEIGHT_UNITY)
                    with dpg.table_row():
                        dpg.add_text("Boukamp (eq. 13)")
                        dpg.add_image(MATH_WEIGHT_BOUKAMP)
                with dpg.group(horizontal=True):
                    dpg.add_image(MATH_Z_FIT)
                    dpg.add_text(
                        "is the real/imaginary part of the ith modeled impedance."
                    )
                with dpg.group(horizontal=True):
                    dpg.add_image(MATH_Z_EXP)
                    dpg.add_text(
                        "is the real/imaginary part of the ith experimental impedance."
                    )
                dpg.add_text(tooltips.fitting.weight, wrap=500)
            self.weight_combo: int = dpg.generate_uuid()
            dpg.add_combo(
                items=list(label_to_weight.keys()),
                default_value="Auto",
                width=-1,
                tag=self.weight_combo,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Max. num. of func. eval.".rjust(label_pad))
            attach_tooltip(tooltips.fitting.nfev)
            self.max_nfev_input: int = dpg.generate_uuid()
            dpg.add_input_int(
                default_value=1000,
                min_value=0,
                min_clamped=True,
                step=0,
                on_enter=True,
                tag=self.max_nfev_input,
                width=-1,
            )
        self.set_settings(default_settings)

    def get_settings(self) -> FitSettings:
        cdc: str = dpg.get_value(self.cdc_input) or ""
        circuit: Optional[Circuit] = dpg.get_item_user_data(self.cdc_input)
        if circuit is None or cdc != circuit.to_string():
            circuit = self.parse_cdc(cdc, self.cdc_input)
        return FitSettings(
            cdc=circuit.to_string(12) if circuit is not None else "",
            method=label_to_cnls_method.get(
                dpg.get_value(self.method_combo), CNLSMethod.AUTO
            ),
            weight=label_to_weight.get(dpg.get_value(self.weight_combo), Weight.AUTO),
            max_nfev=dpg.get_value(self.max_nfev_input),
        )

    def set_settings(self, settings: FitSettings):
        assert type(settings) is FitSettings, settings
        self.parse_cdc(settings.cdc)
        dpg.set_value(self.method_combo, cnls_method_to_label.get(settings.method))
        dpg.set_value(self.weight_combo, weight_to_label.get(settings.weight))
        dpg.set_value(self.max_nfev_input, settings.max_nfev)

    def parse_cdc(self, cdc: str, sender: int = -1) -> Optional[Circuit]:
        assert type(cdc) is str, cdc
        assert type(sender) is int, sender
        try:
            circuit: Circuit = pyimpspec.parse_cdc(cdc)
        except (pyimpspec.ParsingError, pyimpspec.UnexpectedCharacter) as err:
            dpg.bind_item_theme(self.cdc_input, themes.cdc.invalid)
            update_tooltip(self.cdc_tooltip, str(err))
            dpg.show_item(dpg.get_item_parent(self.cdc_tooltip))
            dpg.set_item_user_data(self.cdc_input, None)
            return None
        dpg.bind_item_theme(self.cdc_input, themes.cdc.valid)
        dpg.hide_item(dpg.get_item_parent(self.cdc_tooltip))
        dpg.set_item_user_data(self.cdc_input, circuit)
        if sender != self.cdc_input:
            dpg.set_value(self.cdc_input, circuit.to_string())
        return circuit

    def show_circuit_editor(self):
        if self.circuit_editor is None:
            return
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        dpg.configure_item(
            self.circuit_editor.window,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
        )
        circuit: Optional[Circuit] = None
        try:
            circuit = pyimpspec.parse_cdc(self.get_settings().cdc)
        except pyimpspec.ParsingError:
            pass
        signals.emit(
            Signal.BLOCK_KEYBINDINGS,
            window=self.circuit_editor.window,
            window_object=self.circuit_editor,
        )
        self.circuit_editor.show(circuit)

    def has_active_input(self) -> bool:
        return dpg.is_item_active(self.cdc_input) or dpg.is_item_active(
            self.max_nfev_input
        )


class ParametersTable:
    def __init__(self):
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label=" Parameters",
            leaf=True,
            tag=self._header,
        ):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Element",
                    width_fixed=True,
                )
                attach_tooltip(tooltips.fitting.element)
                dpg.add_table_column(
                    label="Parameter",
                    width_fixed=True,
                )
                attach_tooltip(tooltips.fitting.parameter)
                dpg.add_table_column(
                    label="Value",
                    width_fixed=True,
                )
                attach_tooltip(tooltips.fitting.parameter_value)
                dpg.add_table_column(
                    label="Error (%)",
                )
                attach_tooltip(tooltips.fitting.error)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        dpg.delete_item(self._table, children_only=True, slot=1)

    def populate(self, fit: FitResult):
        dpg.show_item(self._header)
        column_pads: List[int] = [
            10,
            11,
            10,
            10,
        ]
        if fit is None:
            with dpg.table_row(parent=self._table):
                dpg.add_text("".ljust(column_pads[0]))
                dpg.add_text("".ljust(column_pads[1]))
                dpg.add_text("".ljust(column_pads[2]))
                dpg.add_text("".ljust(column_pads[3]))
            return
        element_labels: List[str] = []
        element_tooltips: List[str] = []
        parameter_labels: List[str] = []
        values: List[str] = []
        value_tooltips: List[str] = []
        error_values: List[str] = []
        error_tooltips: List[str] = []
        element_label: str
        element_tooltip: str
        parameter_label: str
        value: str
        value_tooltip: str
        error_value: str
        error_tooltip: str
        parameters: Dict[str, FittedParameter]
        for element in fit.circuit.get_elements():
            Class: Type[Element] = type(element)
            element_label = element.get_label()
            element_tooltip = Class.get_extended_description()
            parameters = fit.parameters[element_label]
            parameter: FittedParameter
            for parameter_label in sorted(parameters):
                parameter = parameters[parameter_label]
                element_labels.append(element_label)
                element_tooltips.append(element_tooltip)
                parameter_labels.append(
                    parameter_label + (" (fixed)" if parameter.fixed else "")
                )
                values.append(
                    f"{format_number(parameter.value, width=9, significants=3)}"
                )
                value_tooltips.append(
                    f"{format_number(parameter.value, decimals=6).strip()}"
                )
                if parameter.stderr is not None:
                    error: float = parameter.get_relative_error() * 100
                    if error > 100.0:
                        error_value = ">100"
                    elif error < 0.01:
                        error_value = "<0.01"
                    else:
                        error_value = (
                            f"{format_number(error, exponent=False, significants=3)}"
                        )
                    error_tooltip = (
                        f"±{format_number(parameter.stderr, decimals=6).strip()}"
                    )
                else:
                    error_value = "-"
                    if not parameter.fixed:
                        error_tooltip = "Unable to estimate error."
                    else:
                        error_tooltip = "Fixed parameter."
                error_values.append(error_value)
                error_tooltips.append(error_tooltip)
        values = align_numbers(values)
        error_values = align_numbers(error_values)
        num_rows: int = 0
        for (
            element_label,
            element_tooltip,
            parameter_label,
            value,
            value_tooltip,
            error_value,
            error_tooltip,
        ) in zip(
            element_labels,
            element_tooltips,
            parameter_labels,
            values,
            value_tooltips,
            error_values,
            error_tooltips,
        ):
            with dpg.table_row(parent=self._table):
                dpg.add_text(element_label.ljust(column_pads[0]))
                if element_tooltip != "":
                    attach_tooltip(element_tooltip)
                dpg.add_text(parameter_label.ljust(column_pads[1]))
                dpg.add_text(value.ljust(column_pads[2]))
                attach_tooltip(value_tooltip)
                dpg.add_text(error_value.ljust(column_pads[3]))
                if error_tooltip != "":
                    attach_tooltip(error_tooltip)
                num_rows += 1
        dpg.set_item_height(self._table, 18 + 23 * max(1, num_rows))


class StatisticsTable:
    def __init__(self):
        label_pad: int = 23
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Statistics", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23 * 9,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Label".rjust(label_pad),
                    width_fixed=True,
                )
                dpg.add_table_column(
                    label="Value",
                    width_fixed=True,
                )
                label: str
                tooltip: str
                for (label, tooltip) in [
                    (
                        "log X²",
                        tooltips.fitting.chisqr,
                    ),
                    (
                        "log X² (reduced)",
                        tooltips.fitting.red_chisqr,
                    ),
                    (
                        "Akaike inf. crit.",
                        tooltips.fitting.aic,
                    ),
                    (
                        "Bayesian inf. crit.",
                        tooltips.fitting.bic,
                    ),
                    (
                        "Degrees of freedom",
                        tooltips.fitting.nfree,
                    ),
                    (
                        "Number of data points",
                        tooltips.fitting.ndata,
                    ),
                    (
                        "Number of func. eval.",
                        tooltips.fitting.nfev,
                    ),
                    (
                        "Method",
                        tooltips.fitting.method,
                    ),
                    (
                        "Weight",
                        tooltips.fitting.weight,
                    ),
                ]:
                    with dpg.table_row():
                        dpg.add_text(label.rjust(label_pad))
                        attach_tooltip(tooltip)
                        tooltip_tag: int = dpg.generate_uuid()
                        dpg.add_text("", user_data=tooltip_tag)
                        attach_tooltip("", tag=tooltip_tag)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            tag: int = dpg.get_item_children(row, slot=1)[2]
            dpg.set_value(tag, "")
            dpg.hide_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))

    def populate(self, fit: FitResult):
        dpg.show_item(self._header)
        cells: List[int] = []
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            cells.append(dpg.get_item_children(row, slot=1)[2])
        assert len(cells) == 9, cells
        tag: int
        value: str
        for (tag, value) in [
            (
                cells[0],
                f"{log(fit.chisqr):.3f}",
            ),
            (
                cells[1],
                f"{log(fit.red_chisqr):.3f}",
            ),
            (
                cells[2],
                f"{fit.aic:.3E}",
            ),
            (
                cells[3],
                f"{fit.bic:.3E}",
            ),
            (
                cells[4],
                f"{fit.nfree}",
            ),
            (
                cells[5],
                f"{fit.ndata}",
            ),
            (
                cells[6],
                f"{fit.nfev}",
            ),
            (
                cells[7],
                cnls_method_to_label.get(fit.method, ""),
            ),
            (
                cells[8],
                weight_to_label.get(fit.weight, ""),
            ),
        ]:
            dpg.set_value(tag, value)
            update_tooltip(dpg.get_item_user_data(tag), value)
            dpg.show_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))
        dpg.set_item_height(self._table, 18 + 23 * 9)


class SettingsTable:
    def __init__(self):
        label_pad: int = 23
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Settings", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Label".rjust(label_pad),
                    width_fixed=True,
                )
                dpg.add_table_column(
                    label="Value",
                    width_fixed=True,
                )
                label: str
                for label in [
                    "Circuit",
                    "Method",
                    "Weight",
                    "Max. num. func. eval.",
                ]:
                    with dpg.table_row():
                        dpg.add_text(label.rjust(label_pad))
                        tooltip_tag: int = dpg.generate_uuid()
                        dpg.add_text("", user_data=tooltip_tag)
                        attach_tooltip("", tag=tooltip_tag)
            with dpg.group(horizontal=True):
                self._apply_settings_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Apply settings",
                    callback=lambda s, a, u: signals.emit(
                        Signal.APPLY_FIT_SETTINGS,
                        **u,
                    ),
                    tag=self._apply_settings_button,
                )
                attach_tooltip(tooltips.general.apply_settings)
                self._apply_mask_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Apply mask",
                    callback=lambda s, a, u: signals.emit(
                        Signal.APPLY_DATA_SET_MASK,
                        **u,
                    ),
                    tag=self._apply_mask_button,
                )
                attach_tooltip(tooltips.general.apply_mask)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            tag: int = dpg.get_item_children(row, slot=1)[1]
            dpg.set_value(tag, "")
            dpg.hide_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))

    def populate(self, fit: FitResult, data: DataSet):
        dpg.show_item(self._header)
        rows: List[int] = []
        cells: List[Tuple[int, int]] = []
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            rows.append(row)
            cells.append(dpg.get_item_children(row, slot=1))
        assert len(rows) == len(cells) == 4, (
            rows,
            cells,
        )
        circuit: Circuit = pyimpspec.parse_cdc(fit.settings.cdc)
        tag: int
        value: str
        for (row, tag, value) in [
            (
                rows[0],
                cells[0][1],
                f"{circuit}\n\n{circuit.to_string(3)}",
            ),
            (
                rows[1],
                cells[1][1],
                cnls_method_to_label.get(fit.settings.method, ""),
            ),
            (
                rows[2],
                cells[2][1],
                weight_to_label.get(fit.settings.weight, ""),
            ),
            (
                rows[3],
                cells[3][1],
                f"{fit.settings.max_nfev}",
            ),
        ]:
            dpg.set_value(tag, value.split("\n")[0])
            update_tooltip(dpg.get_item_user_data(tag), value)
            dpg.show_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))
        dpg.set_item_height(self._table, 18 + 23 * 4)
        dpg.set_item_user_data(
            self._apply_settings_button,
            {"settings": fit.settings},
        )
        dpg.set_item_user_data(
            self._apply_mask_button,
            {
                "data": data,
                "mask": fit.mask,
                "fit": fit,
            },
        )


class DataSetsCombo:
    def __init__(self, label: str, width: int):
        self.labels: List[str] = []
        dpg.add_text(label)
        self.tag: int = dpg.generate_uuid()
        dpg.add_combo(
            callback=lambda s, a, u: signals.emit(
                Signal.SELECT_DATA_SET,
                data=u.get(a),
            ),
            user_data={},
            width=width,
            tag=self.tag,
        )

    def populate(self, labels: List[str], lookup: Dict[str, DataSet]):
        self.labels.clear()
        self.labels.extend(labels)
        label: str = dpg.get_value(self.tag) or ""
        if labels and label not in labels:
            label = labels[0]
        dpg.configure_item(
            self.tag,
            default_value=label,
            items=labels,
            user_data=lookup,
        )

    def get(self) -> Optional[DataSet]:
        return dpg.get_item_user_data(self.tag).get(dpg.get_value(self.tag))

    def set(self, label: str):
        assert type(label) is str, label
        assert label in self.labels, (
            label,
            self.labels,
        )
        dpg.set_value(self.tag, label)

    def clear(self):
        dpg.configure_item(
            self.tag,
            default_value="",
        )


class ResultsCombo:
    def __init__(self, label: str, width: int):
        self.labels: Dict[str, str] = {}
        dpg.add_text(label)
        self.tag: int = dpg.generate_uuid()
        dpg.add_combo(
            callback=lambda s, a, u: signals.emit(
                Signal.SELECT_FIT_RESULT,
                fit=u[0].get(a),
                data=u[1],
            ),
            user_data=(
                {},
                None,
            ),
            width=width,
            tag=self.tag,
        )

    def populate(self, lookup: Dict[str, FitResult], data: Optional[DataSet]):
        self.labels.clear()
        labels: List[str] = list(lookup.keys())
        longest_cdc: int = max(list(map(lambda _: len(_[: _.find(" ")]), labels)) + [1])
        old_key: str
        for old_key in labels:
            fit: FitResult = lookup[old_key]
            del lookup[old_key]
            cdc, timestamp = (
                old_key[: old_key.find(" ")],
                old_key[old_key.find(" ") + 1 :],
            )
            new_key: str = f"{cdc.ljust(longest_cdc)} {timestamp}"
            self.labels[old_key] = new_key
            lookup[new_key] = fit
        labels = list(lookup.keys())
        dpg.configure_item(
            self.tag,
            default_value=labels[0] if labels else "",
            items=labels,
            user_data=(
                lookup,
                data,
            ),
        )

    def get(self) -> Optional[FitResult]:
        return dpg.get_item_user_data(self.tag)[0].get(dpg.get_value(self.tag))

    def set(self, label: str):
        assert type(label) is str, label
        assert label in self.labels, (
            label,
            list(self.labels.keys()),
        )
        dpg.set_value(self.tag, self.labels[label])

    def clear(self):
        dpg.configure_item(
            self.tag,
            default_value="",
        )

    def get_next_result(self) -> Optional[FitResult]:
        lookup: Dict[str, FitResult] = dpg.get_item_user_data(self.tag)[0]
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous_result(self) -> Optional[FitResult]:
        lookup: Dict[str, FitResult] = dpg.get_item_user_data(self.tag)[0]
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) - 1
        return lookup[labels[index % len(labels)]]


class FittingTab:
    def __init__(self, state):
        self.state = state
        self.queued_update: Optional[Callable] = None
        label_pad: int = 24
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(label="Fitting", tag=self.tab):
            with dpg.group(horizontal=True):
                self.sidebar_width: int = 350
                self.sidebar_window: int = dpg.generate_uuid()
                with dpg.child_window(
                    border=False,
                    width=self.sidebar_width,
                    tag=self.sidebar_window,
                ):
                    with dpg.child_window(
                        border=True,
                        width=-1,
                        height=128,
                    ):
                        self.circuit_editor: CircuitEditor = CircuitEditor(
                            window=dpg.add_window(
                                label="Circuit editor",
                                show=False,
                                modal=True,
                                on_close=lambda s, a, u: self.accept_circuit(None),
                            ),
                            callback=self.accept_circuit,
                        )
                        self.settings_menu: SettingsMenu = SettingsMenu(
                            state.config.default_fit_settings,
                            label_pad,
                            circuit_editor=self.circuit_editor,
                        )
                        with dpg.group(horizontal=True):
                            self.visibility_item: int = dpg.generate_uuid()
                            dpg.add_text(
                                "?".rjust(label_pad),
                                tag=self.visibility_item,
                            )
                            attach_tooltip(tooltips.fitting.perform)
                            self.perform_fit_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Perform fit",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.PERFORM_FIT,
                                    data=u,
                                    settings=self.get_settings(),
                                ),
                                user_data=None,
                                width=-1,
                                tag=self.perform_fit_button,
                            )
                    # Results
                    with dpg.child_window(width=-1, height=82):
                        label_pad = 8
                        with dpg.group(horizontal=True):
                            self.data_sets_combo: DataSetsCombo = DataSetsCombo(
                                label="Data set".rjust(label_pad),
                                width=-60,
                            )
                        with dpg.group(horizontal=True):
                            self.results_combo: ResultsCombo = ResultsCombo(
                                label="Result".rjust(label_pad),
                                width=-60,
                            )
                            self.delete_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Delete",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.DELETE_FIT_RESULT,
                                    **u,
                                ),
                                user_data={},
                                width=-1,
                                tag=self.delete_button,
                            )
                            attach_tooltip(tooltips.fitting.delete)
                        with dpg.group(horizontal=True):
                            dpg.add_text("Output".rjust(label_pad))
                            # TODO: Split into combo class?
                            self.output_combo: int = dpg.generate_uuid()
                            dpg.add_combo(
                                items=list(label_to_fit_sim_output.keys()),
                                default_value=list(label_to_fit_sim_output.keys())[0],
                                tag=self.output_combo,
                                width=-60,
                            )
                            self.copy_output_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Copy",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.COPY_OUTPUT,
                                    output=self.get_active_output(),
                                    **u,
                                ),
                                user_data={},
                                width=-1,
                                tag=self.copy_output_button,
                            )
                            attach_tooltip(tooltips.general.copy_output)
                    #
                    with dpg.child_window(width=-1, height=-1):
                        self.result_group: int = dpg.generate_uuid()
                        with dpg.group(tag=self.result_group):
                            with dpg.group(show=False):
                                self.validity_text: int = dpg.generate_uuid()
                                dpg.bind_item_theme(
                                    dpg.add_text(
                                        "",
                                        wrap=self.sidebar_width - 24,
                                        tag=self.validity_text,
                                    ),
                                    themes.result.invalid,
                                )
                                dpg.add_spacer(height=8)
                            self.parameters_table: ParametersTable = ParametersTable()
                            dpg.add_spacer(height=8)
                            self.statistics_table: StatisticsTable = StatisticsTable()
                            dpg.add_spacer(height=8)
                            self.settings_table: SettingsTable = SettingsTable()
                self.plot_window: int = dpg.generate_uuid()
                with dpg.child_window(
                    border=False,
                    width=-1,
                    height=-1,
                    tag=self.plot_window,
                ):
                    self.circuit_preview_height: int = 250
                    with dpg.child_window(
                        border=False,
                        width=-1,
                        height=self.circuit_preview_height,
                    ):
                        dpg.add_text("Fitted circuit")
                        self.circuit_preview: CircuitPreview = CircuitPreview()
                    self.minimum_plot_side: int = 400
                    with dpg.group(horizontal=True):
                        with dpg.group():
                            self.nyquist_plot: Nyquist = Nyquist(
                                width=self.minimum_plot_side,
                                height=self.minimum_plot_side,
                            )
                            self.nyquist_plot.plot(
                                real=array([]),
                                imaginary=array([]),
                                label="Data",
                                theme=themes.nyquist.data,
                            )
                            self.nyquist_plot.plot(
                                real=array([]),
                                imaginary=array([]),
                                label="Fit",
                                fit=True,
                                theme=themes.nyquist.simulation,
                            )
                            self.nyquist_plot.plot(
                                real=array([]),
                                imaginary=array([]),
                                label="Fit",
                                fit=True,
                                line=True,
                                theme=themes.nyquist.simulation,
                                show_label=False,
                            )
                            with dpg.group(horizontal=True):
                                self.enlarge_nyquist_button: int = dpg.generate_uuid()
                                self.adjust_nyquist_limits_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge Nyquist",
                                    callback=self.show_enlarged_nyquist,
                                    tag=self.enlarge_nyquist_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_nyquist_limits_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_nyquist_limits)
                                dpg.add_button(
                                    label="Copy as CSV",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.COPY_PLOT_DATA,
                                        plot=self.nyquist_plot,
                                        context=Context.FITTING_TAB,
                                    ),
                                )
                                attach_tooltip(tooltips.general.copy_plot_data_as_csv)
                        self.horizontal_bode_group: int = dpg.generate_uuid()
                        with dpg.group(tag=self.horizontal_bode_group):
                            self.bode_plot_horizontal: Bode = Bode(
                                width=self.minimum_plot_side,
                                height=self.minimum_plot_side,
                            )
                            self.bode_plot_horizontal.plot(
                                frequency=array([]),
                                magnitude=array([]),
                                phase=array([]),
                                labels=(
                                    "|Z| (d)",
                                    "phi (d)",
                                ),
                                themes=(
                                    themes.bode.magnitude_data,
                                    themes.bode.phase_data,
                                ),
                            )
                            self.bode_plot_horizontal.plot(
                                frequency=array([]),
                                magnitude=array([]),
                                phase=array([]),
                                labels=(
                                    "|Z| (f)",
                                    "phi (f)",
                                ),
                                fit=True,
                                themes=(
                                    themes.bode.magnitude_simulation,
                                    themes.bode.phase_simulation,
                                ),
                            )
                            self.bode_plot_horizontal.plot(
                                frequency=array([]),
                                magnitude=array([]),
                                phase=array([]),
                                labels=(
                                    "|Z| (f)",
                                    "phi (f)",
                                ),
                                fit=True,
                                line=True,
                                themes=(
                                    themes.bode.magnitude_simulation,
                                    themes.bode.phase_simulation,
                                ),
                                show_labels=False,
                            )
                            with dpg.group(horizontal=True):
                                self.enlarge_bode_horizontal_button: int = (
                                    dpg.generate_uuid()
                                )
                                self.adjust_bode_limits_horizontal_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge Bode",
                                    callback=self.show_enlarged_bode,
                                    tag=self.enlarge_bode_horizontal_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_bode_limits_horizontal_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_bode_limits)
                                dpg.add_button(
                                    label="Copy as CSV",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.COPY_PLOT_DATA,
                                        plot=self.bode_plot_horizontal,
                                        context=Context.FITTING_TAB,
                                    ),
                                )
                                attach_tooltip(tooltips.general.copy_plot_data_as_csv)
                    self.vertical_bode_group: int = dpg.generate_uuid()
                    with dpg.group(tag=self.vertical_bode_group):
                        self.bode_plot_vertical: Bode = Bode(
                            width=-1,
                            height=self.minimum_plot_side,
                        )
                        self.bode_plot_vertical.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z| (d)",
                                "phi (d)",
                            ),
                            themes=(
                                themes.bode.magnitude_data,
                                themes.bode.phase_data,
                            ),
                        )
                        self.bode_plot_vertical.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z| (f)",
                                "phi (f)",
                            ),
                            fit=True,
                            themes=(
                                themes.bode.magnitude_simulation,
                                themes.bode.phase_simulation,
                            ),
                        )
                        self.bode_plot_vertical.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z| (f)",
                                "phi (f)",
                            ),
                            fit=True,
                            line=True,
                            themes=(
                                themes.bode.magnitude_simulation,
                                themes.bode.phase_simulation,
                            ),
                            show_labels=False,
                        )
                        with dpg.group(horizontal=True):
                            self.enlarge_bode_vertical_button: int = dpg.generate_uuid()
                            self.adjust_bode_limits_vertical_checkbox: int = (
                                dpg.generate_uuid()
                            )
                            dpg.add_button(
                                label="Enlarge Bode",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.SHOW_ENLARGED_PLOT,
                                    plot=self.bode_plot_vertical,
                                    adjust_limits=dpg.get_value(
                                        self.adjust_bode_limits_horizontal_checkbox
                                    ),
                                ),
                                tag=self.enlarge_bode_vertical_button,
                            )
                            dpg.add_checkbox(
                                default_value=True,
                                source=self.adjust_bode_limits_horizontal_checkbox,
                                tag=self.adjust_bode_limits_vertical_checkbox,
                            )
                            attach_tooltip(tooltips.general.adjust_bode_limits)
                            dpg.add_button(
                                label="Copy as CSV",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.COPY_PLOT_DATA,
                                    plot=self.bode_plot_vertical,
                                    context=Context.FITTING_TAB,
                                ),
                            )
                            attach_tooltip(tooltips.general.copy_plot_data_as_csv)
                    self.residuals_plot: Residuals = Residuals(
                        width=-1,
                        height=300,
                    )
                    self.residuals_plot.plot(
                        frequency=array([]),
                        real=array([]),
                        imaginary=array([]),
                    )
                    with dpg.group(horizontal=True):
                        self.enlarge_residuals_button: int = dpg.generate_uuid()
                        self.adjust_residuals_limits_checkbox: int = dpg.generate_uuid()
                        dpg.add_button(
                            label="Enlarge residuals",
                            callback=self.show_enlarged_residuals,
                            tag=self.enlarge_residuals_button,
                        )
                        dpg.add_checkbox(
                            default_value=True,
                            tag=self.adjust_residuals_limits_checkbox,
                        )
                        attach_tooltip(tooltips.general.adjust_residuals_limits)
                        dpg.add_button(
                            label="Copy as CSV",
                            callback=lambda s, a, u: signals.emit(
                                Signal.COPY_PLOT_DATA,
                                plot=self.residuals_plot,
                                context=Context.FITTING_TAB,
                            ),
                        )
                        attach_tooltip(tooltips.general.copy_plot_data_as_csv)

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self.visibility_item)

    def get_settings(self) -> FitSettings:
        return self.settings_menu.get_settings()

    def set_settings(self, settings: FitSettings):
        self.settings_menu.set_settings(settings)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0
        assert type(height) is int and height > 0
        if not self.is_visible():
            return
        if width < (self.sidebar_width + self.minimum_plot_side * 2):
            if dpg.is_item_shown(self.horizontal_bode_group):
                dpg.hide_item(self.horizontal_bode_group)
                dpg.show_item(self.vertical_bode_group)
                self.nyquist_plot.resize(-1, self.minimum_plot_side)
        else:
            if dpg.is_item_shown(self.vertical_bode_group):
                dpg.show_item(self.horizontal_bode_group)
                dpg.hide_item(self.vertical_bode_group)
            dpg.split_frame()
            width, height = dpg.get_item_rect_size(self.plot_window)
            width = round((width - 24) / 2)
            height = height - 300 - 24 * 2 - 2
            self.nyquist_plot.resize(width, height)
            self.bode_plot_horizontal.resize(width, height)

    def clear(self, hide: bool = True):
        self.data_sets_combo.clear()
        self.results_combo.clear()
        self.parameters_table.clear(hide=hide)
        self.statistics_table.clear(hide=hide)
        self.settings_table.clear(hide=hide)
        self.circuit_preview.clear()
        self.nyquist_plot.clear(delete=False)
        self.bode_plot_horizontal.clear(delete=False)
        self.bode_plot_vertical.clear(delete=False)
        self.residuals_plot.clear(delete=False)

    def populate_data_sets(self, labels: List[str], lookup: Dict[str, DataSet]):
        assert type(labels) is list, labels
        assert type(lookup) is dict, lookup
        self.data_sets_combo.populate(labels, lookup)

    def populate_fits(self, lookup: Dict[str, FitResult], data: Optional[DataSet]):
        assert type(lookup) is dict, lookup
        assert type(data) is DataSet or data is None, data
        self.results_combo.populate(lookup, data)
        dpg.hide_item(dpg.get_item_parent(self.validity_text))
        if data is not None and self.results_combo.labels:
            signals.emit(
                Signal.SELECT_FIT_RESULT,
                fit=self.results_combo.get(),
                data=data,
            )
        else:
            self.parameters_table.clear(hide=True)
            self.statistics_table.clear(hide=True)
            self.settings_table.clear(hide=True)
            self.select_data_set(data)
            dpg.set_item_user_data(
                self.delete_button,
                {},
            )

    def get_next_data_set(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.data_sets_combo.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.data_sets_combo.tag)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous_data_set(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.data_sets_combo.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.data_sets_combo.tag)) - 1
        return lookup[labels[index % len(labels)]]

    def get_next_result(self) -> Optional[FitResult]:
        return self.results_combo.get_next_result()

    def get_previous_result(self) -> Optional[FitResult]:
        return self.results_combo.get_previous_result()

    def select_data_set(self, data: Optional[DataSet]):
        assert type(data) is DataSet or data is None, data
        self.clear(hide=data is None)
        dpg.set_item_user_data(self.perform_fit_button, data)
        if data is None:
            return
        self.data_sets_combo.set(data.get_label())
        real: ndarray
        imag: ndarray
        real, imag = data.get_nyquist_data()
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        freq: ndarray
        mag: ndarray
        phase: ndarray
        freq, mag, phase = data.get_bode_data()
        self.bode_plot_horizontal.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        self.bode_plot_vertical.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )

    def assert_fit_up_to_date(self, fit: FitResult, data: DataSet):
        # Check if the number of unmasked points is the same
        Z_exp: ndarray = data.get_impedance()
        Z_fit: ndarray = fit.get_impedance()
        assert Z_exp.shape == Z_fit.shape, "The number of data points differ!"
        # Check if the masks are the same
        mask_exp: Dict[int, bool] = data.get_mask()
        mask_fit: Dict[int, bool] = {
            k: fit.mask.get(k, mask_exp.get(k, False)) for k in fit.mask
        }
        num_masked_exp: int = list(data.get_mask().values()).count(True)
        num_masked_fit: int = list(fit.mask.values()).count(True)
        assert num_masked_exp == num_masked_fit, "The masks are different sizes!"
        i: int
        for i in mask_fit.keys():
            assert (
                i in mask_exp
            ), f"The data set does not have a point at index {i + 1}!"
            assert (
                mask_exp[i] == mask_fit[i]
            ), f"The data set's mask differs at index {i + 1}!"
        # Check if the frequencies and impedances are the same
        assert allclose(
            fit.get_frequency(), data.get_frequency()
        ), "The frequencies differ!"
        real_residual: ndarray
        imaginary_residual: ndarray
        real_residual, imaginary_residual = _calculate_residuals(Z_exp, Z_fit)
        assert allclose(fit.real_residual, real_residual) and allclose(
            fit.imaginary_residual, imaginary_residual
        ), "The data set's impedances differ from what they were when the fit was performed!"

    def select_fit_result(self, fit: Optional[FitResult], data: Optional[DataSet]):
        assert type(fit) is FitResult or fit is None, fit
        assert type(data) is DataSet or data is None, data
        dpg.set_item_user_data(
            self.delete_button,
            {
                "fit": fit,
                "data": data,
            },
        )
        dpg.set_item_user_data(
            self.copy_output_button,
            {
                "fit_or_sim": fit,
                "data": data,
            },
        )
        if not self.is_visible():
            self.queued_update = lambda: self.select_fit_result(fit, data)
            return
        self.queued_update = None
        self.select_data_set(data)
        if fit is None or data is None:
            if dpg.get_value(self.adjust_nyquist_limits_checkbox):
                self.nyquist_plot.queue_limits_adjustment()
            if dpg.get_value(self.adjust_bode_limits_horizontal_checkbox):
                self.bode_plot_horizontal.queue_limits_adjustment()
                self.bode_plot_vertical.queue_limits_adjustment()
            if dpg.get_value(self.adjust_residuals_limits_checkbox):
                self.residuals_plot.queue_limits_adjustment()
            return
        self.results_combo.set(fit.get_label())
        message: str
        try:
            self.assert_fit_up_to_date(fit, data)
            dpg.hide_item(dpg.get_item_parent(self.validity_text))
        except AssertionError as message:
            dpg.set_value(
                self.validity_text,
                f"Fit result is not valid for the current state of the data set!\n\n{message}",
            )
            dpg.show_item(dpg.get_item_parent(self.validity_text))
        self.parameters_table.populate(fit)
        self.statistics_table.populate(fit)
        self.settings_table.populate(fit, data)
        self.circuit_preview.update(pyimpspec.parse_cdc(fit.settings.cdc))
        real: ndarray
        imag: ndarray
        real, imag = fit.get_nyquist_data()
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        real, imag = fit.get_nyquist_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.nyquist_plot.update(
            index=2,
            real=real,
            imaginary=imag,
        )
        freq: ndarray
        mag: ndarray
        phase: ndarray
        freq, mag, phase = fit.get_bode_data()
        self.bode_plot_horizontal.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = fit.get_bode_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.bode_plot_horizontal.update(
            index=2,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = fit.get_bode_data()
        self.bode_plot_vertical.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = fit.get_bode_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.bode_plot_vertical.update(
            index=2,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, real, imag = fit.get_residual_data()
        self.residuals_plot.update(
            index=0,
            frequency=freq,
            real=real,
            imaginary=imag,
        )
        if dpg.get_value(self.adjust_nyquist_limits_checkbox):
            self.nyquist_plot.queue_limits_adjustment()
        if dpg.get_value(self.adjust_bode_limits_horizontal_checkbox):
            self.bode_plot_horizontal.queue_limits_adjustment()
            self.bode_plot_vertical.queue_limits_adjustment()
        if dpg.get_value(self.adjust_residuals_limits_checkbox):
            self.residuals_plot.queue_limits_adjustment()

    def show_circuit_editor(self):
        self.settings_menu.show_circuit_editor()

    def accept_circuit(self, circuit: Optional[Circuit]):
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)
        self.circuit_editor.hide()
        if circuit is None:
            return
        self.settings_menu.parse_cdc(circuit.to_string(12))

    def show_enlarged_nyquist(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.nyquist_plot,
            adjust_limits=dpg.get_value(self.adjust_nyquist_limits_checkbox),
        )

    def show_enlarged_bode(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.bode_plot_horizontal,
            adjust_limits=dpg.get_value(self.adjust_bode_limits_horizontal_checkbox),
        )

    def show_enlarged_residuals(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.residuals_plot,
            adjust_limits=dpg.get_value(self.adjust_residuals_limits_checkbox),
        )

    def get_active_output(self) -> Optional[FitSimOutput]:
        return label_to_fit_sim_output.get(dpg.get_value(self.output_combo))

    def has_active_input(self) -> bool:
        return self.settings_menu.has_active_input()
